# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json

import pytest
from typer.testing import CliRunner
from unittest.mock import MagicMock, ANY


from dorsal.cli import app

runner = CliRunner()

TEST_DATA_DIR = "tests/data"

# A sample return value for the collection.info() method
MOCK_INFO_RESULT = {
    "by_source": [
        {"source": "cache", "count": 10},
        {"source": "disk", "count": 5},
    ]
}


@pytest.fixture
def mock_build_cache_cmd(mocker):
    """Mocks dependencies for the `cache build` command."""
    # Patch LocalFileCollection at its source due to lazy loading
    mock_collection_class = mocker.patch("dorsal.file.collection.local.LocalFileCollection")

    # Configure the mock instance that the constructor will return
    mock_instance = mock_collection_class.return_value
    mock_instance.__len__.return_value = 15  # 10 from cache + 5 from disk
    mock_instance.info.return_value = MOCK_INFO_RESULT

    return {
        "collection_class": mock_collection_class,
        "collection_instance": mock_instance,
    }


def test_build_cache_default(mock_rich_console, mock_build_cache_cmd):
    """Tests a default cache build, which should run in 'update' mode."""
    # Normalize the path for the CLI runner and the mock assertion
    normalized_path = os.path.normpath(str(TEST_DATA_DIR))
    result = runner.invoke(app, ["cache", "build", normalized_path])

    assert result.exit_code == 0
    # In default mode, it should use the cache to only process new/modified files
    mock_build_cache_cmd["collection_class"].assert_called_once_with(
        source=normalized_path,
        recursive=False,
        console=mock_rich_console,
        palette=ANY,
        use_cache=True,
        follow_symlinks=True,
        lazy=False,
    )

    # Check the formatted text output
    all_printed_text = "".join(str(call.args[0]) for call in mock_rich_console.print.call_args_list)
    assert "Cache build complete" in all_printed_text
    assert "Total files processed: 15" in all_printed_text
    assert "Loaded from cache: 10" in all_printed_text
    assert "Newly added to cache: 5" in all_printed_text


def test_build_cache_force(mock_rich_console, mock_build_cache_cmd):
    """Tests a forced cache build with the --yes flag, which re-processes all files."""
    result = runner.invoke(app, ["cache", "build", TEST_DATA_DIR, "--force"])

    assert result.exit_code == 0
    # With the --force flag, use_cache must be False
    assert mock_build_cache_cmd["collection_class"].call_args.kwargs["use_cache"] is False


def test_build_cache_json_output(mock_rich_console, mock_build_cache_cmd):
    """Tests the --json output flag."""
    result = runner.invoke(app, ["cache", "build", TEST_DATA_DIR, "--json"])

    assert result.exit_code == 0
    mock_rich_console.print.assert_called_once()

    json_str = mock_rich_console.print.call_args.args[0]
    data = json.loads(json_str)

    assert data == {
        "success": True,
        "total_files_processed": 15,
        "loaded_from_cache": 10,
        "newly_added_to_cache": 5,
    }


def test_build_cache_exception_handling(mock_build_cache_cmd):
    """Tests that an exception during collection is handled gracefully."""
    mock_build_cache_cmd["collection_class"].side_effect = Exception("Cannot read directory")

    result = runner.invoke(app, ["cache", "build", TEST_DATA_DIR])

    assert result.exit_code != 0
    assert "An error occurred while building the cache: Cannot read directory" in result.output



# Copyright 2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import MagicMock
import pytest
import typer
from typer.testing import CliRunner


from dorsal.cli.model_app.uninstall_model_cmd import uninstall_model
from dorsal.cli.themes.palettes import DEFAULT_PALETTE
from dorsal.common.exceptions import DorsalError


cli_app = typer.Typer()


@cli_app.callback()
def main_callback(ctx: typer.Context):
    ctx.obj = {"palette": DEFAULT_PALETTE}


cli_app.command(name="uninstall")(uninstall_model)

runner = CliRunner()


@pytest.fixture
def mock_uninstall_cmd(mocker, mock_rich_console):
    """
    Mocks backend dependencies for the `uninstall_model` command.
    """

    mocker.patch("dorsal.common.cli.get_rich_console", return_value=mock_rich_console)

    mock_uninstaller = mocker.patch("dorsal.registry.uninstaller.uninstall_model_target")
    mock_uninstaller.return_value = "dorsal-gpt-neo"

    mock_confirm = mocker.patch("rich.prompt.Confirm.ask", return_value=True)

    return {
        "uninstaller": mock_uninstaller,
        "confirm": mock_confirm,
    }


def test_uninstall_model_interactive_success(mock_rich_console, mock_uninstall_cmd):
    """Tests a standard interactive uninstallation where the user confirms."""
    result = runner.invoke(cli_app, ["uninstall", "dorsal/gpt-neo"])

    assert result.exit_code == 0, result.output

    mock_uninstall_cmd["confirm"].assert_called_once()

    mock_uninstall_cmd["uninstaller"].assert_called_once_with("dorsal/gpt-neo", scope="project")

    assert mock_rich_console.print.called
    panel = mock_rich_console.print.call_args_list[-1].args[0]
    assert "Successfully uninstalled" in panel.renderable
    assert "dorsal-gpt-neo" in panel.renderable
    assert "Uninstall Complete" in str(panel.title)


def test_uninstall_model_interactive_decline(mock_rich_console, mock_uninstall_cmd):
    """Tests that declining the confirmation prompt aborts the operation."""
    mock_uninstall_cmd["confirm"].return_value = False

    result = runner.invoke(cli_app, ["uninstall", "dorsal/gpt-neo"])

    assert result.exit_code == 0, result.output
    assert "Cancelled" in str(mock_rich_console.print.call_args.args[0])

    mock_uninstall_cmd["uninstaller"].assert_not_called()


def test_uninstall_model_yes_flag(mock_uninstall_cmd):
    """Tests that --yes skips the confirmation prompt."""
    result = runner.invoke(cli_app, ["uninstall", "dorsal/gpt-neo", "--yes"])

    assert result.exit_code == 0, result.output
    mock_uninstall_cmd["confirm"].assert_not_called()
    mock_uninstall_cmd["uninstaller"].assert_called_once()


def test_uninstall_model_global_flag(mock_uninstall_cmd):
    """Tests that --global passes the correct scope to the uninstaller."""
    result = runner.invoke(cli_app, ["uninstall", "dorsal/gpt-neo", "--global", "--yes"])

    assert result.exit_code == 0, result.output
    mock_uninstall_cmd["uninstaller"].assert_called_once_with("dorsal/gpt-neo", scope="global")


def test_uninstall_model_dorsal_error(mock_rich_console, mock_uninstall_cmd):
    """Tests handling of specific DorsalErrors (e.g., package not found)."""
    mock_uninstall_cmd["uninstaller"].side_effect = DorsalError("Package not installed.")

    result = runner.invoke(cli_app, ["uninstall", "dorsal/gpt-neo", "--yes"])

    assert result.exit_code != 0

    assert mock_rich_console.print.called
    output_str = str(mock_rich_console.print.call_args.args[0])
    assert "Uninstall Failed" in output_str
    assert "Package not installed" in output_str


def test_uninstall_model_unexpected_error(mock_rich_console, mock_uninstall_cmd):
    """Tests handling of generic unexpected exceptions."""
    mock_uninstall_cmd["uninstaller"].side_effect = Exception("Permission denied")

    result = runner.invoke(cli_app, ["uninstall", "dorsal/gpt-neo", "--yes"])

    assert result.exit_code != 0

    assert mock_rich_console.print.called
    output_str = str(mock_rich_console.print.call_args.args[0])
    assert "Unexpected Error" in output_str
    assert "Permission denied" in output_str



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import json

import pytest
from typer.testing import CliRunner
from unittest.mock import MagicMock
from rich.panel import Panel


from dorsal.cli import app
from dorsal.common.exceptions import DorsalClientError

runner = CliRunner()
COLLECTION_ID = "col_abc123"


class StdinSimulator(io.StringIO):
    def isatty(self):
        return False


@pytest.fixture
def mock_add_files_cmd(mocker):
    """Mocks dependencies for the `collection add-files` command."""
    mock_add_response = MagicMock()
    mock_add_response.added_count = 2
    mock_add_response.duplicate_count = 1
    mock_add_response.invalid_count = 0
    mock_add_response.model_dump_json.return_value = "{}"

    mock_get_collection_response = MagicMock()
    mock_get_collection_response.collection.is_private = True

    mock_add = mocker.patch("dorsal.api.collection.add_files_to_collection", return_value=mock_add_response)
    mock_get = mocker.patch(
        "dorsal.api.collection.get_collection",
        return_value=mock_get_collection_response,
    )

    return {
        "add_files": mock_add,
        "get_collection": mock_get,
        "add_response": mock_add_response,
        "get_collection_response": mock_get_collection_response,
    }


def test_add_files_with_hash_options(mock_rich_console, mock_add_files_cmd):
    """Tests adding files using multiple --hash options."""
    hashes_to_add = ["hash1", "hash2"]
    result = runner.invoke(
        app,
        [
            "collection",
            "add-files",
            COLLECTION_ID,
            "--hash",
            hashes_to_add[0],
            "--hash",
            hashes_to_add[1],
        ],
    )

    assert result.exit_code == 0
    # The command may change the order, so we check the contents
    called_hashes = mock_add_files_cmd["add_files"].call_args.kwargs["hashes"]
    assert sorted(called_hashes) == sorted(hashes_to_add)

    # Check for the success panel
    printed_object = mock_rich_console.print.call_args.args[0]
    assert isinstance(printed_object, Panel)
    assert "Update Complete" in str(printed_object.title)


def test_add_files_from_stdin(mock_rich_console, mock_add_files_cmd, mocker):
    """Tests adding files by reading hashes from stdin."""
    hashes_to_add = ["hash1", "hash2", "hash3"]

    mocker.patch(
        "dorsal.cli.collection_app.add_files_cmd._get_hashes",
        return_value=hashes_to_add,
    )

    result = runner.invoke(
        app,
        [
            "collection",
            "add-files",
            COLLECTION_ID,
            "--from-stdin",
        ],
    )

    assert result.exit_code == 0
    called_hashes = mock_add_files_cmd["add_files"].call_args.kwargs["hashes"]
    assert sorted(called_hashes) == sorted(hashes_to_add)


def test_add_files_no_hashes_error():
    """Tests that the command fails if no hashes are provided."""
    result = runner.invoke(app, ["collection", "add-files", COLLECTION_ID])

    assert result.exit_code != 0
    assert "No file hashes provided" in result.output


def test_add_files_json_output(mock_rich_console, mock_add_files_cmd):
    """Tests a successful add with --json output."""
    result = runner.invoke(app, ["collection", "add-files", COLLECTION_ID, "--hash", "hash1", "--json"])

    assert result.exit_code == 0
    mock_add_files_cmd["add_response"].model_dump_json.assert_called_once()
    mock_rich_console.print.assert_called_once_with(mock_add_files_cmd["add_response"].model_dump_json.return_value)


def test_add_files_with_invalid_count(mock_rich_console, mock_add_files_cmd):
    """Tests that the command fetches collection info when some hashes are invalid."""
    mock_add_files_cmd["add_response"].invalid_count = 1

    result = runner.invoke(app, ["collection", "add-files", COLLECTION_ID, "--hash", "hash1"])

    assert result.exit_code == 0

    mock_add_files_cmd["get_collection"].assert_called_once_with(
        collection_id=COLLECTION_ID, hydrate=False, mode="pydantic"
    )

    printed_object = mock_rich_console.print.call_args.args[0]
    assert isinstance(printed_object, Panel)
    panel_text = str(printed_object.renderable)
    assert "Not Added: 1" in panel_text
    assert "Note: Files can only be added" in panel_text



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

MEDIAINFO_DISPLAY_FIELDS = {
    # Track types
    "Audio",
    "Menu",
    "General",
    "Other",
    "Image",
    "Video",
    "Text",
    # Misc
    "url",
    # --- GENERAL / TOP-LEVEL FIELDS ---
    "UniqueID_String",
    "FileSize_String",
    "Duration_String",
    "OverallBitRate_String",
    "FrameRate_String",
    "Encoded_Date",
    "Tagged_Date",
    "Encoded_Application_String",
    "Encoded_Library_String",
    # --- COMMON METADATA ---
    "Title",
    "Album",
    "Performer",
    "Composer",
    "Director",
    "Genre",
    "Comment",
    "Description",
    "Synopsis",
    "Copyright",
    # --- CORE FORMAT & CODEC FIELDS (for all tracks) ---
    "ID",
    "Format_Version",
    "Format_String",
    "Format_Info",
    "Format_Profile",
    "CodecID",
    "CodecID_Info",
    "InternetMediaType",
    "Compression_Mode_String",
    # --- DETAILED ENCODING SETTINGS (primarily for Video) ---
    "Encoded_Library_Settings",
    "Format_Settings",
    "Format_Settings_CABAC_String",
    "Format_Settings_RefFrames_String",
    "Format_Settings_GOP",
    "BitsPixel_Frame",
    # --- VIDEO-SPECIFIC FIELDS ---
    "Width",
    "Height",
    "DisplayAspectRatio_String",
    "FrameRate_Mode_String",
    "ScanType_String",
    "HDR_Format_String",
    # --- AUDIO-SPECIFIC FIELDS ---
    "Channels_String",
    "ChannelLayout",
    "SamplingRate_String",
    # --- IMAGE/COLOR-SPECIFIC FIELDS ---
    "ColorSpace",
    "ChromaSubsampling",
    "BitDepth_String",
    "colour_range",  # Note: UK spelling from MediaInfo
    "colour_primaries",
    "transfer_characteristics",
    "matrix_coefficients",
    "Format_Compression",
    # --- SUBTITLE-SPECIFIC FIELDS ---
    "ElementCount",
    # --- STREAM-LEVEL FIELDS ---
    "BitRate_String",
    "BitRate_Mode_String",
    "StreamSize_String",
    "Language_String",
    "Delay_String",
    "Default_String",
    "Forced_String",
    # --- OTHER / TIMECODE FIELDS ---
    "Type",
    "TimeCode_FirstFrame",
    "TimeCode_LastFrame",
    "TimeCode_Striped_String",
    # --- Fields to Exclude ---
    # --- These are either redundant, too verbose, or not typically shown ---
    # Note: The goal behind selection/exclusion here is to provide similar display fields to the MediaInfo desktop tool
    # "Accompaniment",
    # "ActiveFormatDescription",
    # "Actor",
    # "Actor_Character",
    # "Added_Date",
    # "Album_More",
    # "Album_Performer",
    # "Album_Performer_Sort",
    # "Album_Performer_Url",
    # "Album_ReplayGain_Gain",
    # "Album_ReplayGain_Gain_String",
    # "Album_ReplayGain_Peak",
    # "Album_Sort",
    # "Alignment",
    # "Alignment_String",
    # "AlternateGroup",
    # "AlternateGroup_String",
    # "Archival_Location",
    # "Arranger",
    # "ArtDirector",
    # "AssistantDirector",
    # "Audio_Channels_Total",
    # "Audio_Codec_List",
    # "AudioCount",
    # "Audio_Format_List",
    # "Audio_Format_WithHint_List",
    # "Audio_Language_List",
    # "BarCode",
    # "BitDepth", # Redundant: BitDepth_String is used
    # "BitDepth_Detected",
    # "BitDepth_Detected_String",
    # "BitDepth_Stored",
    # "BitDepth_Stored_String",
    # "BitRate", # Redundant: BitRate_String is used
    # "BitRate_Encoded",
    # "BitRate_Encoded_String",
    # "BitRate_Maximum",
    # "BitRate_Maximum_String",
    # "BitRate_Minimum",
    # "BitRate_Minimum_String",
    # "BitRate_Mode", # Redundant: BitRate_Mode_String is used
    # "BitRate_Nominal",
    # "BitRate_Nominal_String",
    # "Bits-Pixel_Frame", # Alias for BitsPixel_Frame
    # "BPM",
    # "BufferSize",
    # "CatalogNumber",
    # "ChannelLayout_Original",
    # "ChannelLayoutID",
    # "ChannelPositions",
    # "ChannelPositions_Original",
    # "ChannelPositions_Original_String2",
    # "ChannelPositions_String2",
    # "Channels", # Redundant: Channels_String is used
    # "Channels_Original",
    # "Channels_Original_String",
    # "Chapter",
    # "Chapters_Pos_Begin",
    # "Chapters_Pos_End",
    # "Choregrapher",
    # "ChromaSubsampling_Position",
    # "ChromaSubsampling_String", # Redundant: ChromaSubsampling is cleaner
    # "Codec",
    # "Codec_CC",
    # "Codec_Description",
    # "Codec_Extensions",
    # "Codec_Family",
    # "CodecID_Compatible",
    # "CodecID_Description",
    # "CodecID_Hint",
    # "CodecID_String",
    # "CodecID_Url",
    # "CodecID_Version",
    # "Codec_Info", # Redundant: Format_Info is generally better
    # "Codec_Profile",
    # "Codec_Settings", # Too generic, specific settings are used
    # "Codec_Settings_Automatic",
    # "Codec_Settings_BVOP",
    # "Codec_Settings_CABAC", # Redundant: Codec_Settings_CABAC_String is used
    # "Codec_Settings_Endianness",
    # "Codec_Settings_Firm",
    # "Codec_Settings_Floor",
    # "Codec_Settings_GMC",
    # "Codec_Settings_GMC_String",
    # "Codec_Settings_ITU",
    # "Codec_Settings_Law",
    # "Codec_Settings_Matrix",
    # "Codec_Settings_Matrix_Data",
    # "Codec_Settings_PacketBitStream",
    # "Codec_Settings_QPel",
    # "Codec_Settings_RefFrames", # Redundant: Codec_Settings_RefFrames_String is used
    # "Codec_Settings_Sign",
    # "Codec_Url",
    # "CoDirector",
    # "Collection",
    # "Colorimetry",
    # "colour_description_present",
    # "colour_description_present_Original",
    # "colour_description_present_Original_Source",
    # "colour_description_present_Source",
    # "colour_primaries_Original",
    # "colour_primaries_Original_Source",
    # "colour_primaries_Source",
    # "colour_range_Original",
    # "colour_range_Original_Source",
    # "colour_range_Source",
    # "Comic",
    # "Comic_More",
    # "Comic_Position_Total",
    # "CommissionedBy",
    # "Compilation",
    # "Compilation_String",
    # "Composer_Nationality",
    # "Composer_Sort",
    # "Compression_Mode", # Redundant: Compression_Mode_String is used
    # "Compression_Ratio",
    # "Conductor",
    # "ContentType",
    # "CoProducer",
    # "Copyright_Url",
    # "CostumeDesigner",
    # "Count",
    # "Countries",
    # "Country",
    # "Cover",
    # "Cover_Data",
    # "Cover_Description",
    # "Cover_Mime",
    # "Cover_Type",
    # "Cropped",
    # "DataSize",
    # "Default", # Redundant: Default_String is used
    # "Delay", # Redundant: Delay_String is used
    # "Delay_DropFrame",
    # "Delay_Original",
    # "Delay_Original_DropFrame",
    # "Delay_Original_Settings",
    # "Delay_Original_Source",
    # "Delay_Original_String",
    # "Delay_Original_String1",
    # "Delay_Original_String2",
    # "Delay_Original_String3",
    # "Delay_Original_String4",
    # "Delay_Original_String5",
    # "Delay_Settings",
    # "Delay_Source",
    # "Delay_Source_String",
    # "Delay_String1",
    # "Delay_String2",
    # "Delay_String3",
    # "Delay_String4",
    # "Delay_String5",
    # "Dimensions",
    # "DirectorOfPhotography",
    # "Disabled",
    # "Disabled_String",
    # "DisplayAspectRatio", # Redundant: DisplayAspectRatio_String is used
    # "DisplayAspectRatio_CleanAperture",
    # "DisplayAspectRatio_CleanAperture_String",
    # "DisplayAspectRatio_Original",
    # "DisplayAspectRatio_Original_String",
    # "DistributedBy",
    # "DolbyVision_Layers",
    # "DolbyVision_Profile",
    # "DolbyVision_Version",
    # "Domain",
    # "DotsPerInch",
    # "Duration", # Redundant: Duration_String is used
    # "Duration_Base",
    # "Duration_End",
    # "Duration_End_Command",
    # "Duration_End_Command_String",
    # "Duration_End_Command_String1",
    # "Duration_End_Command_String2",
    # "Duration_End_Command_String3",
    # "Duration_End_Command_String4",
    # "Duration_End_Command_String5",
    # "Duration_End_String",
    # "Duration_End_String1",
    # "Duration_End_String2",
    # "Duration_End_String3",
    # "Duration_End_String4",
    # "Duration_End_String5",
    # "Duration_FirstFrame",
    # "Duration_FirstFrame_String",
    # "Duration_FirstFrame_String1",
    # "Duration_FirstFrame_String2",
    # "Duration_FirstFrame_String3",
    # "Duration_FirstFrame_String4",
    # "Duration_FirstFrame_String5",
    # "Duration_LastFrame",
    # "Duration_LastFrame_String",
    # "Duration_LastFrame_String1",
    # "Duration_LastFrame_String2",
    # "Duration_LastFrame_String3",
    # "Duration_LastFrame_String4",
    # "Duration_LastFrame_String5",
    # "Duration_Start",
    # "Duration_Start2End",
    # "Duration_Start2End_String",
    # "Duration_Start2End_String1",
    # "Duration_Start2End_String2",
    # "Duration_Start2End_String3",
    # "Duration_Start2End_String4",
    # "Duration_Start2End_String5",
    # "Duration_Start_Command",
    # "Duration_Start_Command_String",
    # "Duration_Start_Command_String1",
    # "Duration_Start_Command_String2",
    # "Duration_Start_Command_String3",
    # "Duration_Start_Command_String4",
    # "Duration_Start_Command_String5",
    # "Duration_Start_String",
    # "Duration_Start_String1",
    # "Duration_Start_String2",
    # "Duration_Start_String3",
    # "Duration_Start_String4",
    # "Duration_Start_String5",
    # "Duration_String1",
    # "Duration_String2",
    # "Duration_String3",
    # "Duration_String4",
    # "Duration_String5",
    # "EditedBy",
    # "Encoded_Application", # Redundant: Encoded_Application_String is used
    # "Encoded_Application_CompanyName",
    # "Encoded_Application_Name",
    # "Encoded_Application_Url",
    # "Encoded_Application_Version",
    # "EncodedBy",
    # "Encoded_Library", # Redundant: Encoded_Library_String is used
    # "Encoded_Library_CompanyName",
    # "Encoded_Library_Date",
    # "Encoded_Library_Name",
    # "Encoded_Library_Settings", # Redundant: Use specific settings fields
    # "Encoded_Library_Version",
    # "Encoded_OperatingSystem",
    # "Encryption",
    # "Encryption_Format",
    # "Encryption_InitializationVector",
    # "Encryption_Length",
    # "Encryption_Method",
    # "Encryption_Mode",
    # "Encryption_Padding",
    # "EPG_Positions_Begin",
    # "EPG_Positions_End",
    # "Events_MinDuration",
    # "Events_MinDuration_String",
    # "Events_MinDuration_String1",
    # "Events_MinDuration_String2",
    # "Events_MinDuration_String3",
    # "Events_MinDuration_String4",
    # "Events_MinDuration_String5",
    # "Events_PaintOn",
    # "Events_PopOn",
    # "Events_RollUp",
    # "Events_Total",
    # "ExecutiveProducer",
    # "extra",
    # "File_Created_Date",
    # "File_Created_Date_Local",
    # "FileExtension",
    # "FileExtension_Last",
    # "File_Modified_Date",
    # "File_Modified_Date_Local",
    # "FileName",
    # "FileNameExtension",
    # "FileNameExtension_Last",
    # "FileName_Last",
    # "FileSize", # Redundant: FileSize_String is used
    # "FileSize_String1",
    # "FileSize_String2",
    # "FileSize_String3",
    # "FileSize_String4",
    # "FirstDisplay_Delay_Frames",
    # "FirstDisplay_Type",
    # "FirstPacketOrder",
    # "FooterSize",
    # "Forced", # Redundant: Forced_String is used
    # "Format_AdditionalFeatures",
    # "Format_Commercial",
    # "Format_Commercial_IfAny",
    # "Format_Extensions",
    # "Format_Level",
    # "Format_Settings_BVOP",
    # "Format_Settings_BVOP_String",
    # "Format_Settings_Emphasis",
    # "Format_Settings_Endianness",
    # "Format_Settings_Firm",
    # "Format_Settings_Floor",
    # "Format_Settings_FrameMode",
    # "Format_Settings_GMC",
    # "Format_Settings_GMC_String",
    # "Format_Settings_ITU",
    # "Format_Settings_Law",
    # "Format_Settings_Matrix",
    # "Format_Settings_Matrix_Data",
    # "Format_Settings_Matrix_String",
    # "Format_Settings_Mode",
    # "Format_Settings_ModeExtension",
    # "Format_Settings_Packing",
    # "Format_Settings_PictureStructure",
    # "Format_Settings_PS",
    # "Format_Settings_PS_String",
    # "Format_Settings_Pulldown",
    # "Format_Settings_QPel",
    # "Format_Settings_QPel_String",
    # "Format_Settings_SBR",
    # "Format_Settings_SBR_String",
    # "Format_Settings_Sign",
    # "Format_Settings_SliceCount",
    # "Format_Settings_SliceCount_String",
    # "Format_Settings_Wrapping",
    # "Format_Tier",
    # "Format_Url",
    # "FrameCount",
    # "FrameRate", # Redundant: FrameRate_String is used
    # "FrameRate_Den",
    # "FrameRate_Maximum",
    # "FrameRate_Maximum_String",
    # "FrameRate_Minimum",
    # "FrameRate_Minimum_String",
    # "FrameRate_Mode", # Redundant: FrameRate_Mode_String is used
    # "FrameRate_Mode_Original",
    # "FrameRate_Mode_Original_String",
    # "FrameRate_Nominal",
    # "FrameRate_Nominal_String",
    # "FrameRate_Num",
    # "FrameRate_Original",
    # "FrameRate_Original_Den",
    # "FrameRate_Original_Num",
    # "FrameRate_Original_String",
    # "FrameRate_Real",
    # "FrameRate_Real_String",
    # "GeneralCount",
    # "Gop_OpenClosed",
    # "Gop_OpenClosed_FirstFrame",
    # "Gop_OpenClosed_FirstFrame_String",
    # "Gop_OpenClosed_String",
    # "Grouping",
    # "HDR_Format_Commercial",
    # "HDR_Format_Compatibility",
    # "HDR_Format_Level",
    # "HDR_Format_Profile",
    # "HDR_Format_Settings",
    # "HDR_Format_Version",
    # "HeaderSize",
    # "Height_CleanAperture",
    # "Height_CleanAperture_String",
    # "Height_Offset",
    # "Height_Offset_String",
    # "Height_Original",
    # "Height_Original_String",
    # "Height_String", # Redundant: Height (raw value) is used
    # "ICRA",
    # "ID_String", # Redundant: ID is used
    # "Image_Codec_List",
    # "ImageCount",
    # "Image_Format_List",
    # "Image_Format_WithHint_List",
    # "Image_Language_List",
    # "Inform",
    # "Interlacement",
    # "Interlacement_String",
    # "Interleaved",
    # "Interleave_Duration",
    # "Interleave_Duration_String",
    # "Interleave_Preload",
    # "Interleave_Preload_String",
    # "Interleave_VideoFrames",
    # "ISBN",
    # "ISRC",
    # "IsStreamable",
    # "Keywords",
    # "Label",
    # "LabelCode",
    # "Language", # Redundant: Language_String is used
    # "Language_More",
    # "Language_String1",
    # "Language_String2",
    # "Language_String3",
    # "Language_String4",
    # "LawRating",
    # "LawRating_Reason",
    # "LCCN",
    # "Lightness",
    # "Lines_Count",
    # "Lines_MaxCharacterCount",
    # "Lines_MaxCountPerEvent",
    # "List",
    # "List_StreamKind",
    # "List_StreamPos",
    # "List_String",
    # "Lyricist",
    # "Lyrics",
    # "MasteredBy",
    # "Mastered_Date",
    # "MasteringDisplay_ColorPrimaries",
    # "MasteringDisplay_ColorPrimaries_Original",
    # "MasteringDisplay_ColorPrimaries_Original_Source",
    # "MasteringDisplay_ColorPrimaries_Source",
    # "MasteringDisplay_Luminance",
    # "MasteringDisplay_Luminance_Original",
    # "MasteringDisplay_Luminance_Original_Source",
    # "MasteringDisplay_Luminance_Source",
    # "matrix_coefficients_Original",
    # "matrix_coefficients_Original_Source",
    # "matrix_coefficients_Source",
    # "Matrix_ChannelPositions",
    # "Matrix_ChannelPositions_String2",
    # "Matrix_Channels",
    # "Matrix_Channels_String",
    # "Matrix_Format",
    # "MaxCLL",
    # "MaxCLL_Original",
    # "MaxCLL_Original_Source",
    # "MaxCLL_Source",
    # "MaxFALL",
    # "MaxFALL_Original",
    # "MaxFALL_Original_Source",
    # "MaxFALL_Source",
    # "Menu_Codec_List",
    # "MenuCount",
    # "Menu_Format_List",
    # "Menu_Format_WithHint_List",
    # "MenuID",
    # "MenuID_String",
    # "Menu_Language_List",
    # "Mood",
    # "Movie",
    # "Movie_Country",
    # "Movie_More",
    # "Movie_Url",
    # "MultiView_BaseProfile",
    # "MultiView_Count",
    # "MultiView_Layout",
    # "MusicBy",
    # "MuxingMode",
    # "MuxingMode_MoreInfo",
    # "NetworkName",
    # "Original_Album",
    # "Original_Lyricist",
    # "Original_Movie",
    # "Original_NetworkName",
    # "OriginalNetworkName",
    # "Original_Part",
    # "Original_Performer",
    # "Original_Released_Date",
    # "OriginalSourceForm",
    # "OriginalSourceForm_Cropped",
    # "OriginalSourceForm_DistributedBy",
    # "OriginalSourceForm_Name",
    # "OriginalSourceForm_NumColors",
    # "OriginalSourceForm_Sharpness",
    # "OriginalSourceMedium",
    # "OriginalSourceMedium_ID",
    # "OriginalSourceMedium_ID_String",
    # "Original_Track",
    # "Other_Codec_List",
    # "OtherCount",
    # "Other_Format_List",
    # "Other_Format_WithHint_List",
    # "Other_Language_List",
    # "OverallBitRate", # Redundant: OverallBitRate_String is used
    # "OverallBitRate_Maximum",
    # "OverallBitRate_Maximum_String",
    # "OverallBitRate_Minimum",
    # "OverallBitRate_Minimum_String",
    # "OverallBitRate_Mode",
    # "OverallBitRate_Mode_String",
    # "OverallBitRate_Nominal",
    # "OverallBitRate_Nominal_String",
    # "Owner",
    # "PackageName",
    # "Part",
    # "Part_Position",
    # "Part_Position_Total",
    # "Performer_Sort",
    # "Performer_Url",
    # "Period",
    # "PixelAspectRatio",
    # "PixelAspectRatio_CleanAperture",
    # "PixelAspectRatio_CleanAperture_String",
    # "PixelAspectRatio_Original",
    # "PixelAspectRatio_Original_String",
    # "PixelAspectRatio_String",
    # "Played_Count",
    # "Played_First_Date",
    # "Played_Last_Date",
    # "PodcastCategory",
    # "Producer",
    # "Producer_Copyright",
    # "ProductionDesigner",
    # "ProductionStudio",
    # "Publisher",
    # "Publisher_URL",
    # "Rating",
    # "Reel",
    # "Reel_Position",
    # "Reel_Position_Total",
    # "RemixedBy",
    # "ReplayGain_Gain",
    # "ReplayGain_Gain_String",
    # "ReplayGain_Peak",
    # "Resolution",
    # "Resolution_String",
    # "Rotation",
    # "Rotation_String",
    # "Sampled_Height",
    # "Sampled_Width",
    # "SamplesPerFrame",
    # "SamplingCount",
    # "SamplingRate", # Redundant: SamplingRate_String is used
    # "ScanOrder",
    # "ScanOrder_Original",
    # "ScanOrder_Original_String",
    # "ScanOrder_Stored",
    # "ScanOrder_StoredDisplayedInverted",
    # "ScanOrder_Stored_String",
    # "ScanOrder_String",
    # "ScanType", # Redundant: ScanType_String is used
    # "ScanType_Original",
    # "ScanType_Original_String",
    # "ScanType_StoreMethod",
    # "ScanType_StoreMethod_FieldsPerBlock",
    # "ScanType_StoreMethod_String",
    # "ScreenplayBy",
    # "Season",
    # "Season_Position",
    # "Season_Position_Total",
    # "ServiceChannel",
    # "ServiceKind",
    # "ServiceKind_String",
    # "ServiceName",
    # "ServiceProvider",
    # "ServiceProvider_Url",
    # "ServiceType",
    # "Service_Url",
    # "SoundEngineer",
    # "Source_Duration",
    # "Source_Duration_FirstFrame",
    # "Source_Duration_FirstFrame_String",
    # "Source_Duration_FirstFrame_String1",
    # "Source_Duration_FirstFrame_String2",
    # "Source_Duration_FirstFrame_String3",
    # "Source_Duration_FirstFrame_String4",
    # "Source_Duration_FirstFrame_String5",
    # "Source_Duration_LastFrame",
    # "Source_Duration_LastFrame_String",
    # "Source_Duration_LastFrame_String1",
    # "Source_Duration_LastFrame_String2",
    # "Source_Duration_LastFrame_String3",
    # "Source_Duration_LastFrame_String4",
    # "Source_Duration_LastFrame_String5",
    # "Source_Duration_String",
    # "Source_Duration_String1",
    # "Source_Duration_String2",
    # "Source_Duration_String3",
    # "Source_Duration_String4",
    # "Source_Duration_String5",
    # "Source_FrameCount",
    # "Source_SamplingCount",
    # "Source_StreamSize",
    # "Source_StreamSize_Encoded",
    # "Source_StreamSize_Encoded_Proportion",
    # "Source_StreamSize_Encoded_String",
    # "Source_StreamSize_Encoded_String1",
    # "Source_StreamSize_Encoded_String2",
    # "Source_StreamSize_Encoded_String3",
    # "Source_StreamSize_Encoded_String4",
    # "Source_StreamSize_Encoded_String5",
    # "Source_StreamSize_Proportion",
    # "Source_StreamSize_String",
    # "Source_StreamSize_String1",
    # "Source_StreamSize_String2",
    # "Source_StreamSize_String3",
    # "Source_StreamSize_String4",
    # "Source_StreamSize_String5",
    # "Standard",
    # "Status",
    # "Stored_Height",
    # "Stored_Width",
    # "StreamCount",
    # "StreamKind",
    # "StreamKindID",
    # "StreamKindPos",
    # "StreamKind_String",
    # "StreamOrder",
    # "StreamSize", # Redundant: StreamSize_String is used
    # "StreamSize_Demuxed",
    # "StreamSize_Demuxed_String",
    # "StreamSize_Demuxed_String1",
    # "StreamSize_Demuxed_String2",
    # "StreamSize_Demuxed_String3",
    # "StreamSize_Demuxed_String4",
    # "StreamSize_Demuxed_String5",
    # "StreamSize_Encoded",
    # "StreamSize_Encoded_Proportion",
    # "StreamSize_Encoded_String",
    # "StreamSize_Encoded_String1",
    # "StreamSize_Encoded_String2",
    # "StreamSize_Encoded_String3",
    # "StreamSize_Encoded_String4",
    # "StreamSize_Encoded_String5",
    # "StreamSize_Proportion",
    # "StreamSize_String1",
    # "StreamSize_String2",
    # "StreamSize_String3",
    # "StreamSize_String4",
    # "StreamSize_String5",
    # "Subject",
    # "SubTrack",
    # "Summary",
    # "Tagged_Application",
    # "TermsOfUse",
    # "Text_Codec_List",
    # "TextCount",
    # "Text_Format_List",
    # "Text_Format_WithHint_List",
    # "Text_Language_List",
    # "ThanksTo",
    # "TimeCode_DropFrame",
    # "TimeCode_MaxFrameNumber",
    # "TimeCode_MaxFrameNumber_Theory",
    # "TimeCode_Settings",
    # "TimeCode_Source",
    # "TimeCode_Striped", # Redundant: TimeCode_Striped_String is used
    # "TimeCode_Stripped",
    # "TimeCode_Stripped_String",
    # "TimeStamp_FirstFrame",
    # "TimeStamp_FirstFrame_String",
    # "TimeStamp_FirstFrame_String1",
    # "TimeStamp_FirstFrame_String2",
    # "TimeStamp_FirstFrame_String3",
    # "TimeStamp_FirstFrame_String4",
    # "TimeStamp_FirstFrame_String5",
    # "TimeZone",
    # "TimeZones",
    # "Title_More",
    # "Title_Url",
    # "Track",
    # "Track_More",
    # "Track_Position",
    # "Track_Position_Total",
    # "Track_Sort",
    # "Track_Url",
    # "transfer_characteristics_Original",
    # "transfer_characteristics_Original_Source",
    # "transfer_characteristics_Source",
    # "UMID",
    # "UniqueID", # Redundant: UniqueID_String is used
    # "UniversalAdID_Registry",
    # "UniversalAdID_String",
    # "UniversalAdID_Value",
    # "Video0_Delay",
    # "Video0_Delay_String",
    # "Video0_Delay_String1",
    # "Video0_Delay_String2",
    # "Video0_Delay_String3",
    # "Video0_Delay_String4",
    # "Video0_Delay_String5",
    # "Video_Codec_List",
    # "VideoCount",
    # "Video_Delay",
    # "Video_Delay_String",
    # "Video_Delay_String1",
    # "Video_Delay_String2",
    # "Video_Delay_String3",
    # "Video_Delay_String4",
    # "Video_Delay_String5",
    # "Video_Format_List",
    # "Video_Format_WithHint_List",
    # "Video_Language_List",
    # "Width_CleanAperture",
    # "Width_CleanAperture_String",
    # "Width_Offset",
    # "Width_Offset_String",
    # "Width_Original",
    # "Width_Original_String",
    # "Width_String", # Redundant: Width (raw value) is used
    # "WrittenBy",
    # "Written_Location",
}



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from dorsal.common.validators.json_schema import JsonSchemaValidator

    audio_transcription: JsonSchemaValidator
    classification: JsonSchemaValidator
    document_extraction: JsonSchemaValidator
    embedding: JsonSchemaValidator
    entity_extraction: JsonSchemaValidator
    generic: JsonSchemaValidator
    geolocation: JsonSchemaValidator
    llm_output: JsonSchemaValidator
    object_detection: JsonSchemaValidator
    regression: JsonSchemaValidator


def __getattr__(name: str):
    """
    Lazy-loads attributes from the 'open_schema' submodule.
    This ensures that simply importing 'dorsal.file.validators' is instant
    and does not trigger the heavy schema parsing machinery.
    """
    from dorsal.file.validators import open_schema

    try:
        return getattr(open_schema, name)
    except AttributeError as err:
        raise AttributeError(f"module '{__name__}' has no attribute '{name}'") from err


def __dir__() -> List[str]:
    """
    Exposes the available validators to dir(), enabling autocomplete in
    interactive shells (IPython/Jupyter) and dynamic inspection.
    """
    from dorsal.file.validators import open_schema

    return sorted(list(globals().keys()) + list(open_schema.__all__))



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import zipfile
import xml.etree.ElementTree as ET
from typing import Any


from .utils import (
    RELS_NS,
    CORE_PROPS_TYPE,
    APP_PROPS_TYPE,
    DEFAULT_CORE_PROPS_PATH,
    DEFAULT_APP_PROPS_PATH,
    BASE_NS_MAP,
    _find_relationship_target,
    _read_xml_from_zip,
    _parse_core_properties,
    _parse_app_properties,
    _parse_content_types_xml,
    _parse_custom_properties_xml,
    _find_xml_element,
    _findall_xml_elements,
)
from dorsal.common.language import normalize_language_name, extract_locale_code, normalize_language_alpha3

logger = logging.getLogger(__name__)

W_NS = "{http://schemas.openxmlformats.org/wordprocessingml/2006/main}"

NS_MAP = BASE_NS_MAP.copy()
NS_MAP.update(
    {
        "w": "http://schemas.openxmlformats.org/wordprocessingml/2006/main",
    }
)

HYPERLINK_TYPE = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/hyperlink"
IMAGE_TYPE = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/image"
COMMENTS_TYPE = "http://schemas.openxmlformats.org/wordprocessingml/2006/relationships/comments"


def _parse_settings_xml(xml_root: ET.Element | None) -> dict[str, Any]:
    """Extracts language and track_changes from settings.xml."""
    if xml_root is None:
        return {}

    lang_el = _find_xml_element(xml_root, "w:lang", NS_MAP)
    if lang_el is None:
        lang_el = _find_xml_element(xml_root, "w:themeFontLang", NS_MAP)

    raw_lang_str = lang_el.get(f"{W_NS}val") if lang_el is not None else None
    track_changes_el = _find_xml_element(xml_root, "w:trackRevisions", NS_MAP)

    return {
        "language": normalize_language_name(raw_lang_str),
        "language_code": normalize_language_alpha3(raw_lang_str),
        "locale_code": extract_locale_code(raw_lang_str),
        "has_track_changes": (track_changes_el is not None),
    }


def _parse_styles_xml(xml_root: ET.Element | None) -> dict[str, Any]:
    """Extracts default font from styles.xml."""
    font_el = _find_xml_element(xml_root, "w:docDefaults/w:rPrDefault/w:rPr/w:rFonts", NS_MAP)
    font_text = font_el.get(f"{W_NS}ascii") if font_el is not None else None
    return {"default_font": font_text}


def _parse_font_table_xml(xml_root: ET.Element | None) -> dict[str, Any]:
    """Extracts all declared fonts from fontTable.xml."""
    if xml_root is None:
        return {"all_fonts": []}
    fonts = []
    for el in _findall_xml_elements(xml_root, "w:font", NS_MAP):
        font_name = el.get(f"{W_NS}name")
        if font_name:
            fonts.append(font_name)
    return {"all_fonts": sorted(list(set(fonts)))}


def _parse_document_rels_xml(xml_root: ET.Element | None) -> dict[str, Any]:
    """Extracts hyperlinks and embedded image paths from document.xml.rels."""
    if xml_root is None:
        return {"hyperlinks": [], "embedded_images": []}
    hyperlinks = []
    images = []
    for el in xml_root.findall(f"{RELS_NS}Relationship"):
        rel_type = el.get("Type")
        if rel_type == HYPERLINK_TYPE:
            target = el.get("Target")
            if target:
                hyperlinks.append(target)
        elif rel_type == IMAGE_TYPE:
            target = el.get("Target")
            if target:
                images.append(target)
    return {"hyperlinks": sorted(list(set(hyperlinks))), "embedded_images": sorted(list(set(images)))}


def extract_docx_metadata(file_path: str) -> dict[str, Any] | None:
    """
    Extracts metadata from a .docx file and returns it as a dictionary.
    """
    final_data: dict[str, Any] = {}
    word_specific_data: dict[str, Any] = {}
    is_password_protected = False

    core_props_path: str | None = None
    app_props_path: str | None = None

    try:
        with zipfile.ZipFile(file_path, "r") as z:
            rels_root = _read_xml_from_zip(z, "_rels/.rels")
            if rels_root is None:
                logger.warning("File is not a valid OOXML (missing _rels/.rels): '%s'", file_path)
                return None

            core_props_path = _find_relationship_target(rels_root, CORE_PROPS_TYPE)
            app_props_path = _find_relationship_target(rels_root, APP_PROPS_TYPE)

            if core_props_path is None:
                core_props_path = DEFAULT_CORE_PROPS_PATH
            if app_props_path is None:
                app_props_path = DEFAULT_APP_PROPS_PATH

            core_root = _read_xml_from_zip(z, core_props_path)
            app_root = _read_xml_from_zip(z, app_props_path)
            ct_root = _read_xml_from_zip(z, "[Content_Types].xml")
            custom_props_root = _read_xml_from_zip(z, "docProps/custom.xml")

            settings_root = _read_xml_from_zip(z, "word/settings.xml")
            styles_root = _read_xml_from_zip(z, "word/styles.xml")
            font_table_root = _read_xml_from_zip(z, "word/fontTable.xml")
            doc_rels_root = _read_xml_from_zip(z, "word/_rels/document.xml.rels")

            core_data = _parse_core_properties(core_root, NS_MAP)
            app_data, word_data = _parse_app_properties(app_root, NS_MAP)
            ct_data = _parse_content_types_xml(ct_root)
            custom_data = _parse_custom_properties_xml(custom_props_root, NS_MAP)

            settings_data = _parse_settings_xml(settings_root)
            styles_data = _parse_styles_xml(styles_root)
            font_data = _parse_font_table_xml(font_table_root)
            rels_data = _parse_document_rels_xml(doc_rels_root)

            final_data.update(core_data)
            final_data.update(app_data)
            final_data.update(ct_data)
            final_data.update(settings_data)
            final_data.update(styles_data)
            final_data.update(font_data)
            final_data.update(custom_data)

            word_specific_data.update(word_data)
            word_specific_data.update(rels_data)

            if "has_track_changes" in final_data:
                word_specific_data["has_track_changes"] = final_data.pop("has_track_changes")

    except zipfile.BadZipFile:
        logger.warning("File is not a valid ZIP (corrupt, password-protected, or owner file): '%s'", file_path)
        is_password_protected = True
    except Exception as e:
        logger.error(f"Unexpected error opening docx zip file '{file_path}': {e}", exc_info=True)
        return None

    final_data["is_password_protected"] = is_password_protected

    final_data["excel"] = None
    final_data["powerpoint"] = None
    final_data["word"] = word_specific_data

    found_data_keys = [
        k
        for k, v in final_data.items()
        if (v not in [None, [], {}]) and (k not in ["is_password_protected", "word", "excel", "powerpoint"])
    ]
    found_word_keys = [k for k, v in word_specific_data.items() if v not in [None, [], {}]]

    if not found_data_keys and not found_word_keys:
        if is_password_protected:
            return {"is_password_protected": True}

        logger.warning("Could not extract any metadata from '%s'. File is likely minimal or empty.", file_path)
        return None

    return final_data



# Copyright 2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
from unittest.mock import MagicMock
import pytest
import typer
from typer.testing import CliRunner


from dorsal.cli.model_app.init_model_cmd import init_model
from dorsal.cli.themes.palettes import DEFAULT_PALETTE
from dorsal.common.exceptions import DorsalError


cli_app = typer.Typer()


@cli_app.callback()
def main_callback(ctx: typer.Context):
    ctx.obj = {"palette": DEFAULT_PALETTE}


cli_app.command(name="init")(init_model)

runner = CliRunner()


@pytest.fixture
def mock_init_cmd(mocker, mock_rich_console):
    """
    Mocks backend dependencies for the `init_model` command.
    """

    mocker.patch("dorsal.common.cli.get_rich_console", return_value=mock_rich_console)

    mock_create_project = mocker.patch("dorsal.registry.initialize.create_new_annotation_model_project")

    mock_result = MagicMock()
    mock_result.clean_name = "receipt-scanner"
    mock_result.path = pathlib.Path("/tmp/receipt-scanner")
    mock_create_project.return_value = mock_result

    return {
        "create_project": mock_create_project,
        "result": mock_result,
    }


def test_init_model_success_default_dir(mock_rich_console, mock_init_cmd):
    """Tests initializing a model in the current directory (default)."""
    result = runner.invoke(cli_app, ["init", "Receipt Scanner"])

    assert result.exit_code == 0, result.output

    mock_init_cmd["create_project"].assert_called_once_with(name="Receipt Scanner", target_dir=None)

    assert mock_rich_console.print.called
    panel = mock_rich_console.print.call_args.args[0]
    assert "Created new model project" in panel.renderable
    assert "receipt-scanner" in panel.renderable
    assert "Model Initialized" in str(panel.title)


def test_init_model_success_explicit_dir(mock_rich_console, mock_init_cmd):
    """Tests initializing a model with a specific target directory."""
    with runner.isolated_filesystem():
        target_path = pathlib.Path("my_projects")
        target_path.mkdir()

        result = runner.invoke(cli_app, ["init", "Receipt Scanner", str(target_path)])

        assert result.exit_code == 0, result.output

        args = mock_init_cmd["create_project"].call_args
        assert args.kwargs["name"] == "Receipt Scanner"
        assert args.kwargs["target_dir"].name == "my_projects"


def test_init_model_dorsal_error(mock_rich_console, mock_init_cmd):
    """Tests handling of specific DorsalErrors (e.g., directory already exists)."""
    mock_init_cmd["create_project"].side_effect = DorsalError("Project directory already exists.")

    result = runner.invoke(cli_app, ["init", "Receipt Scanner"])

    assert result.exit_code != 0

    assert mock_rich_console.print.called
    output_str = str(mock_rich_console.print.call_args.args[0])
    assert "Failed to create project" in output_str
    assert "Project directory already exists" in output_str


def test_init_model_unexpected_error(mock_rich_console, mock_init_cmd):
    """Tests handling of generic unexpected exceptions."""
    mock_init_cmd["create_project"].side_effect = Exception("Disk full")

    result = runner.invoke(cli_app, ["init", "Receipt Scanner"])

    assert result.exit_code != 0

    assert mock_rich_console.print.called
    output_str = str(mock_rich_console.print.call_args.args[0])
    assert "Unexpected Error" in output_str
    assert "Disk full" in output_str



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from functools import cached_property
import logging
import os
import pathlib
from typing import Any, Iterable, Literal, Sequence, Type, TYPE_CHECKING, overload

import requests
from tqdm import tqdm
from rich.progress import (
    Progress,
    SpinnerColumn,
    BarColumn,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
    TaskProgressColumn,
    MofNCompleteColumn,
)
from rich.console import Console


from dorsal.version import __version__
from dorsal.client import DorsalClient
from dorsal.client.validators import FileIndexResponse
from dorsal.common import constants
from dorsal.common.auth import is_offline_mode
from dorsal.common.environment import is_jupyter_environment, should_show_progress
from dorsal.common.exceptions import (
    BatchSizeError,
    DorsalError,
    DorsalClientError,
    DuplicateFileError,
)
from dorsal.file.cache import DorsalCache
from dorsal.file.dorsal_file import LocalFile
from dorsal.file.model_runner import ModelRunner
from dorsal.file.utils.files import get_file_paths
from dorsal.session import get_shared_dorsal_client, get_shared_cache


if TYPE_CHECKING:
    from dorsal.file.validators.file_record import FileRecordStrict

logger = logging.getLogger(__name__)


def make_dorsalhub_file_url(file_hash: str) -> str:
    """Constructs a URL to a file record on DorsalHub."""
    base = constants.BASE_URL.rstrip("/")
    f_hash = file_hash.strip("/")
    return f"{base}/file/{f_hash}"


class MetadataReader:
    """A high-level utility for processing local files and indexing metadata.

    This class provides a convenient interface to the `ModelRunner` engine.
    It is designed to simplify the process of scanning local files, extracting
    rich metadata based on a configurable pipeline, and optionally pushing
    the results directly to a DorsalHub instance.

    Example:
        ```python
        from dorsal.file import MetadataReader

        # Initialize the reader
        reader = MetadataReader()

        # Use the reader to process a directory and get LocalFile objects
        local_files = reader.scan_directory("path/to/my_data")
        print(f"Processed {len(local_files)} files.")

        # Or, use the reader to process and immediately index a directory
        summary = reader.index_directory("path/to/my_data")
        print(f"Indexed {summary['success']} files to DorsalHub.")
        ```

    Args:
        client (DorsalClient, optional): An existing `DorsalClient` instance
            to use for API operations. If not provided, a shared instance
            will be used. Defaults to None.
        model_config (str | list, optional): A path to a custom JSON
            pipeline configuration file or a dictionary defining the pipeline
            for the ModelRunner. If None, the default pipeline is used.
            Defaults to None.
    """

    identity = f"dorsal.file.MetadataReader-{__version__}"

    def __init__(
        self,
        api_key: str | None = None,
        client: DorsalClient | None = None,
        model_config: str | list[dict[str, Any]] | None = "default",
        ignore_duplicates: bool = False,
        base_url: str = constants.BASE_URL,
        file_class: Type[LocalFile] = LocalFile,
        offline: bool = False,
    ):
        """
        Args:
            api_key: Optional API key for DorsalHub. If not provided, DorsalClient
                     will attempt to read it from environment variables.
            model_config: Optional configuration for the ModelRunner instance used for
                          local file processing.
            ignore_duplicates: If True, duplicate files (based on content hash)
                               encountered during directory indexing will be skipped.
                               If False (default), a DuplicateFileError will be raised.
            base_url: The base URL for the Dorsal API. Defaults to the common BASE_URL.
            file_class: Currently only `LocalFile` supported.
        """
        self.offline = offline or is_offline_mode()
        self._ignore_duplicates = ignore_duplicates
        self._model_runner = ModelRunner(pipeline_config=model_config)

        if self.offline:
            if client is not None:
                logger.warning("MetadataReader initialized in OFFLINE mode. The provided 'client' will be ignored.")
            self._client_instance = None
        else:
            self._client_instance = client

        self._api_key = api_key
        self._base_url = base_url
        self._file_class = file_class
        logger.debug("MetadataReader initialized.")

    @property
    def _cache(self) -> DorsalCache:
        """Dynamically retrieve the current valid shared cache."""
        return get_shared_cache()

    @property
    def _client(self) -> DorsalClient:
        """Lazily initializes/returns the DorsalClient instance.

        Priority:
        1. Use `self._client_instance` if already set (e.g., passed via __init__).
        2. If `self._api_key` is set, create a new, dedicated client and store it.
        3. If neither is set, fall back to the global shared client.

        Note: when 'passing on' the client, use `_client_instance` instead, as invoking this
              necessarily creates a DorsalClient or raises an error.
        """
        if self.offline:
            raise DorsalError("MetadataReader is in OFFLINE mode. Network operations are blocked.")
        if self._client_instance is None:
            if self._api_key is not None:
                logger.debug("Creating DorsalClient instance using provided api_key.")
                self._client_instance = DorsalClient(api_key=self._api_key)
            else:
                logger.debug("DorsalClient not present, getting shared instance...")
                self._client_instance = get_shared_dorsal_client(api_key=None)

        return self._client_instance

    def _run_models(self, file_path: str, follow_symlinks: bool = True) -> FileRecordStrict:
        """
        Internal helper to run local file models on a given file path.

        Args:
            file_path: Path to the file to process.

        Returns:
            A FileRecordStrict object with generated metadata.

        Raises:
            FileNotFoundError: If the file_path does not exist.
            IOError: If the file cannot be read.
            DorsalClientError: For other errors during model execution, wrapped for consistency.
        """
        try:
            return self._model_runner.run(file_path=file_path, follow_symlinks=follow_symlinks)
        except FileNotFoundError:
            logger.error("File not found by ModelRunner: %s", file_path)
            raise
        except IOError as e:
            logger.error("IOError processing file %s: %s", file_path, e)
            raise
        except Exception as e:
            logger.error("Error running models on file %s: %s", file_path, e)
            raise DorsalClientError(
                message=f"Failed to process metadata for file: {file_path}.",
                original_exception=e,
            ) from e

    def _get_or_create_record(
        self, file_path: str, *, skip_cache: bool, overwrite_cache: bool = False, follow_symlinks: bool = True
    ) -> FileRecordStrict:
        """
        Gets a file record from cache or creates it by running the ModelRunner.

        Args:
            file_path: Path to the file to process.
            skip_cache: If True, the cache check is bypassed and the ModelRunner is forced to run.

        Returns:
            FileRecordStrict object with generated metadata.

        Raises:
            FileNotFoundError: If the file_path does not exist.
            IOError: If the file cannot be read.
            DorsalClientError: For other errors during model execution.
        """
        from dorsal.file.validators.file_record import FileRecordStrict

        if not skip_cache and not overwrite_cache:
            abspath = os.path.abspath(file_path)
            try:
                modified_time = os.path.getmtime(abspath)
                cached_record = self._cache.get_record(path=abspath)

                if cached_record and cached_record.modified_time == modified_time:
                    logger.debug("Cache hit for file: %s", abspath)
                    record = FileRecordStrict.model_validate_json(cached_record.record_json)
                    record.source = "cache"
                    return record

            except FileNotFoundError:
                logger.error("File not found during cache check: %s", file_path)
                raise
            except Exception:
                logger.exception(
                    "Unexpected error during cache retrieval for '%s'. Proceeding to run models.",
                    file_path,
                )

        logger.debug("Cache miss or skipped for file: %s. Running models.", file_path)
        try:
            file_record = self._model_runner.run(file_path=file_path, follow_symlinks=follow_symlinks)

            if not skip_cache or overwrite_cache:
                abspath = os.path.abspath(file_path)
                modified_time = os.path.getmtime(abspath)
                self._cache.upsert_record(path=abspath, modified_time=modified_time, record=file_record)

            return file_record

        except FileNotFoundError:
            logger.error("File not found by ModelRunner: %s", file_path)
            raise
        except IOError as err:
            logger.error("IOError processing file %s: %s", file_path, err)
            raise
        except Exception as err:
            logger.error("Error running models on file %s: %s", file_path, err, exc_info=True)
            raise DorsalClientError(
                message=f"Failed to process metadata for file: {file_path}.",
                original_exception=err,
            ) from err

    def generate_processed_records_from_directory(
        self,
        dir_path: str,
        *,
        recursive: bool = False,
        limit: int | None = None,
        console: Console | None = None,
        show_progress: bool | None = None,
        palette: dict | None = None,
        skip_cache: bool = False,
        overwrite_cache: bool = False,
        follow_symlinks: bool = True,
        lazy: bool = False,
    ) -> tuple[list[FileRecordStrict], dict[str, str]]:
        """
        Scan directory, and sends each file through the ModelRunner pipeline.

        NOTE: When duplicates are ignored, this returned map intentionally contains only the path of the *first* file
              encountered for each hash

        Args:
            dir_path: Path to the directory to scan.
            recursive: If True, scans subdirectories recursively. Defaults to False.
            limit: Optional. top processing files once this many unique records have been generated.
            lazy: If True, consumes file paths via an iterator (so doesn't count files ahead of processing).

        Returns:
            tuple[list[FileRecordStrict], dict[str, str]]:
                - list of unique `FileRecordStrict` objects.
                - mapping file content hashes to original file paths.

        Raises:
            FileNotFoundError: If `dir_path` does not exist.
            DuplicateFileError: If `self._ignore_duplicates` is False and duplicates found.
            DorsalClientError: For errors scanning directory or during local processing.
        """
        logger.debug(
            "Generating processed records from directory: '%s' (Recursive: %s, Ignore Duplicates: %s, Max: %s, Lazy: %s)",
            dir_path,
            recursive,
            self._ignore_duplicates,
            limit,
            lazy,
        )
        try:
            file_paths = get_file_paths(dir_path=dir_path, recursive=recursive, lazy=lazy)
        except FileNotFoundError:
            logger.error("Directory not found for generating records: %s", dir_path)
            raise
        except Exception as e:
            logger.error("Error scanning directory '%s' for record generation: %s", dir_path, e)
            raise DorsalClientError(
                message=f"An error occurred while scanning directory: {dir_path}.",
                original_exception=e,
            ) from e

        total_count = None
        if isinstance(file_paths, list):
            if not file_paths:
                logger.debug("No files found to process in directory: %s", dir_path)
                return [], {}
            if limit and len(file_paths) > limit:
                error_message = f"Number of records in the directory exceeds the limit: {len(file_paths)} > {limit}"
                raise DorsalError(error_message)
            total_count = len(file_paths)

            logger.debug(
                "Generating metadata for %s file(s) from directory '%s' (Max: %s).",
                total_count,
                dir_path,
                limit if limit is not None else "All",
            )
        else:
            logger.debug(
                "Generating metadata from directory '%s' in LAZY mode (Max: %s).",
                dir_path,
                limit if limit is not None else "All",
            )

        file_hash_to_path_map: dict[str, str] = {}
        records_to_index_list: list[FileRecordStrict] = []
        processed_files_summary_local: dict[str, str] = {}

        should_display = should_show_progress(show_progress, console)

        rich_progress = None
        iterator: Iterable[str]

        if should_display and is_jupyter_environment():
            iterator = tqdm(file_paths, desc="Generating Metadata (Local)", total=total_count)
        elif should_display:
            from dorsal.cli.themes.palettes import DEFAULT_PALETTE

            active_palette = palette if palette is not None else DEFAULT_PALETTE
            progress_columns = (
                SpinnerColumn(),
                TextColumn(
                    "{task.description}",
                    style=active_palette.get("progress_description", "default"),
                ),
                BarColumn(complete_style=active_palette.get("progress_bar", "default")),
                TaskProgressColumn(style=active_palette.get("progress_percentage", "default")),
                MofNCompleteColumn(),
                TimeElapsedColumn(),
                TimeRemainingColumn(),
            )
            rich_progress = Progress(
                *progress_columns,
                console=console,
                redirect_stdout=True,
                redirect_stderr=True,
                transient=True,
            )
            task_id = rich_progress.add_task("Initializing...", total=total_count)
            iterator = file_paths
        else:
            iterator = file_paths

        with rich_progress if rich_progress else open(os.devnull, "w"):
            for file_path in iterator:
                if rich_progress:
                    filename = os.path.basename(file_path)
                    rich_progress.update(task_id, description=f"Processing {filename}...")

                if limit is not None and len(records_to_index_list) >= limit:
                    logger.debug(
                        "Reached generation limit of %d records. Stopping further local processing for '%s'.",
                        limit,
                        dir_path,
                    )
                    break
                try:
                    processing_path = file_path
                    if follow_symlinks:
                        try:
                            path_obj = pathlib.Path(file_path)
                            resolved = path_obj.resolve()
                            if resolved.exists():
                                processing_path = str(resolved)
                        except (OSError, RuntimeError) as err:
                            logger.debug("Failed to resolve symlink: %s - %s", file_path, err)
                            pass
                    file_record = self._get_or_create_record(
                        file_path=processing_path, skip_cache=skip_cache, overwrite_cache=overwrite_cache
                    )
                    if not hasattr(file_record, "hash") or not file_record.hash:
                        logger.error(
                            "Skipping file %s: processed record is missing a hash.",
                            file_path,
                        )
                        processed_files_summary_local[file_path] = "error_missing_hash"
                        continue
                    if file_record.hash in file_hash_to_path_map:
                        original_file = file_hash_to_path_map[file_record.hash]
                        if self._ignore_duplicates:
                            logger.debug(
                                "Skipping duplicate: '%s' (original: '%s', hash: %s)",
                                file_path,
                                original_file,
                                file_record.hash,
                            )
                            processed_files_summary_local[file_path] = "skipped_duplicate_content"
                            continue
                        else:
                            logger.error(
                                "Duplicate file content: '%s' and '%s' (hash: %s). Not ignoring.",
                                file_path,
                                original_file,
                                file_record.hash,
                            )
                            raise DuplicateFileError(
                                message="Duplicate file content detected during local processing.",
                                file_paths=[file_path, original_file],
                            )
                    file_hash_to_path_map[file_record.hash] = file_path
                    records_to_index_list.append(file_record)
                    processed_files_summary_local[file_path] = "processed_for_indexing"
                except DuplicateFileError:
                    raise
                except (FileNotFoundError, IOError) as err:
                    logger.error("Skipping '%s': local access error: %s", file_path, err)
                    processed_files_summary_local[file_path] = f"error_local_access: {type(err).__name__}"
                except DorsalClientError as err:
                    logger.error("Skipping '%s': model execution error: %s", file_path, err)
                    processed_files_summary_local[file_path] = (
                        f"error_model_execution: {type(err.original_exception or err).__name__}"
                    )
                except Exception as err:
                    logger.exception(
                        "Unexpected error processing file '%s'. Skipping this file.",
                        file_path,
                    )
                    processed_files_summary_local[file_path] = f"error_unexpected: {type(err).__name__}"

                if rich_progress:
                    rich_progress.advance(task_id)

        logger.debug(
            "Completed local generation for '%s'. %d unique records generated.",
            dir_path,
            len(records_to_index_list),
        )
        return records_to_index_list, file_hash_to_path_map

    def upload_records(
        self,
        records: list[FileRecordStrict],
        *,
        public: bool = False,
        fail_fast: bool = True,
        hash_to_path_map: dict[str, str] | None = None,
        console: "Console | None" = None,
        show_progress: bool | None = None,
        palette: dict | None = None,
    ) -> dict[str, Any]:
        """
        Uploads a list of file records to DorsalHub in batches.

        Args:
            records: List of validated FileRecordStrict objects to upload.
            public: If True, uploads to the public index.
            fail_fast: If True, raises BatchIndexingError on the first failed batch.
            hash_to_path_map: Optional mapping of {hash: local_path}.
            console: Rich Console instance for progress reporting.
            palette: Theme palette for the progress bar.

        Returns:
            dict: A summary of the upload operation.
        """
        from dorsal.common.exceptions import BatchIndexingError
        from dorsal.client.validators import FileIndexResponse
        from dorsal.common.environment import is_jupyter_environment
        from rich.progress import (
            Progress,
            SpinnerColumn,
            BarColumn,
            TaskProgressColumn,
            MofNCompleteColumn,
            TextColumn,
            TimeElapsedColumn,
            TimeRemainingColumn,
        )

        total_records = len(records)
        summary: dict[str, Any] = {
            "total_records": total_records,
            "processed": 0,
            "success": 0,
            "failed": 0,
            "batches": [],
            "errors": [],
        }

        if not records:
            return summary

        batches = [
            records[i : i + constants.API_MAX_BATCH_SIZE] for i in range(0, total_records, constants.API_MAX_BATCH_SIZE)
        ]

        logger.debug(
            "Uploading %d records in %d batches (Public: %s).",
            total_records,
            len(batches),
            public,
        )

        should_display = should_show_progress(show_progress, console)

        rich_progress = None
        iterator: Iterable[list[FileRecordStrict]]

        if should_display and is_jupyter_environment():
            from tqdm import tqdm

            iterator = tqdm(batches, desc="Pushing batches")
        elif should_display:
            from dorsal.cli.themes.palettes import DEFAULT_PALETTE

            active_palette = palette if palette is not None else DEFAULT_PALETTE

            progress_columns = (
                SpinnerColumn(),
                TextColumn(
                    "{task.description}",
                    style=active_palette.get("progress_description", "default"),
                ),
                BarColumn(complete_style=active_palette.get("progress_bar", "default")),
                TaskProgressColumn(style=active_palette.get("progress_percentage", "default")),
                MofNCompleteColumn(),
                TimeElapsedColumn(),
                TimeRemainingColumn(),
            )
            rich_progress = Progress(
                *progress_columns,
                console=console,
                redirect_stdout=True,
                transient=True,
                redirect_stderr=True,
            )
            task_id = rich_progress.add_task("Initializing...", total=len(batches))
            iterator = batches
        else:
            iterator = batches

        with rich_progress if rich_progress else open(os.devnull, "w"):
            for i, batch in enumerate(iterator):
                batch_size = len(batch)
                batch_num = i + 1

                if rich_progress:
                    rich_progress.update(task_id, description=f"Pushing batch {batch_num}/{len(batches)}...")

                try:
                    api_response: FileIndexResponse
                    if public:
                        api_response = self._client.index_public_file_records(file_records=batch)
                    else:
                        api_response = self._client.index_private_file_records(file_records=batch)

                    if api_response.results:
                        for result_item in api_response.results:
                            if hash_to_path_map and hasattr(result_item, "hash") and hasattr(result_item, "file_path"):
                                local_path = hash_to_path_map.get(result_item.hash)
                                if local_path:
                                    result_item.file_path = local_path

                            if hasattr(result_item, "annotations") and result_item.annotations:
                                for ann in result_item.annotations:
                                    if hasattr(ann, "status") and ann.status in ("error", "unauthorized"):
                                        summary["errors"].append(
                                            {
                                                "batch_index": i,
                                                "status": ann.status,
                                                "file_hash": result_item.hash,
                                                "file_path": getattr(result_item, "file_path", None),
                                                "error_message": f"Annotation '{ann.name}': {ann.detail or 'Unknown error'}",
                                            }
                                        )

                    summary["success"] += api_response.success
                    summary["processed"] += batch_size
                    summary["batches"].append(
                        {
                            "batch_index": i,
                            "status": "success",
                            "records_in_batch": batch_size,
                            "records_accepted": api_response.success,
                        }
                    )
                    logger.debug("Batch %d/%d succeeded.", batch_num, len(batches))

                except Exception as err:
                    summary["failed"] += batch_size
                    summary["processed"] += batch_size
                    error_entry = {
                        "batch_index": i,
                        "status": "failure",
                        "records_in_batch": batch_size,
                        "error_type": type(err).__name__,
                        "error_message": str(err),
                    }
                    summary["batches"].append(error_entry)
                    summary["errors"].append(error_entry)
                    logger.error("Batch %d failed: %s", batch_num, err)

                    if fail_fast:
                        raise BatchIndexingError(
                            f"Batch indexing failed at batch {batch_num}: {err}",
                            summary=summary,
                            original_error=err,
                        ) from err

                if rich_progress:
                    rich_progress.advance(task_id)

        return summary

    def index_directory(
        self,
        dir_path: str,
        *,
        recursive: bool = False,
        public: bool = False,
        skip_cache: bool = False,
        fail_fast: bool = True,
        console: "Console | None" = None,
        show_progress: bool | None = None,
        palette: dict | None = None,
        follow_symlinks: bool = True,
        lazy: bool = False,
    ) -> dict:
        """Scans, processes, and indexes all files in a directory to DorsalHub.

        This method performs a complete workflow:
        1. Scans the directory for files.
        2. Runs the metadata extraction pipeline on each unique file.
        3. Indexes the records to DorsalHub in managed batches.

        This method supports two modes of operation via `fail_fast`:
        - **True (Default):** Aborts immediately if a batch fails. Useful for
          interactive sessions or debugging.
        - **False:** Logs errors and continues processing subsequent batches.
          Useful for bulk uploads where partial success is acceptable.

        Example:
            ```python
            from dorsal.file import MetadataReader
            from dorsal.common.exceptions import BatchIndexingError

            reader = MetadataReader()

            # Scenario 1: Interactive / Fail-Fast (Default)
            try:
                summary = reader.index_directory("path/to/data")
                print(f"Success! {summary['success']} records indexed.")
            except BatchIndexingError as e:
                print(f"Aborted after batch {e.summary['processed']}. Error: {e}")

            # Scenario 2: Bulk Upload / Best-Effort
            summary = reader.index_directory(
                "path/to/large_dataset",
                fail_fast=False
            )
            print(f"Finished. Success: {summary['success']}, Failed: {summary['failed']}")
            if summary['errors']:
                print(f"First error: {summary['errors'][0]['error_message']}")
            ```

        Args:
            dir_path (str): The path to the directory to scan and index.
            recursive (bool, optional): If True, scans all subdirectories.
            public (bool, optional): If True, records are created as public.
            skip_cache (bool, optional): If True, re-processes files locally
                even if they haven't changed.
            fail_fast (bool, optional): If True, raises `BatchIndexingError`
                immediately on batch failure. If False, records the error
                and continues. Defaults to True.
            lazy: If True, consumes file paths via an iterator (so doesn't count files ahead of processing).

        Returns:
            dict: A summary dictionary detailing the results.
                  Keys: 'total_records', 'processed', 'success', 'failed', 'batches', 'errors'.

        Raises:
            FileNotFoundError: If `dir_path` does not exist.
            BatchIndexingError: If a batch fails and `fail_fast` is True.
            DorsalClientError: For critical errors before batching begins.
        """
        from dorsal.common.exceptions import DorsalClientError, DuplicateFileError

        logger.debug(
            "MetadataReader.index_directory called: dir='%s', rec=%s, public=%s, fail_fast=%s",
            dir_path,
            recursive,
            public,
            fail_fast,
        )
        if self.offline:
            raise DorsalError("Cannot index directory: MetadataReader is in OFFLINE mode.")

        records_to_index_list, file_hash_to_path_map = self.generate_processed_records_from_directory(
            dir_path=dir_path,
            recursive=recursive,
            limit=None,
            skip_cache=skip_cache,
            console=console,
            show_progress=show_progress,
            palette=palette,
            follow_symlinks=follow_symlinks,
            lazy=lazy,
        )

        if not records_to_index_list:
            return {
                "total_records": 0,
                "processed": 0,
                "success": 0,
                "failed": 0,
                "batches": [],
                "errors": [],
            }

        return self.upload_records(
            records=records_to_index_list,
            public=public,
            fail_fast=fail_fast,
            hash_to_path_map=file_hash_to_path_map,
            console=console,
            palette=palette,
        )

    def index_file(
        self, file_path: str, *, public: bool = False, skip_cache: bool = False, overwrite_cache: bool = False
    ) -> FileIndexResponse:
        """Processes a single file and immediately indexes it to DorsalHub.

        This method provides a complete "read and push" workflow for a single
        file. It runs the metadata extraction pipeline and then uploads the
        resulting record to DorsalHub.

        Example:
            ```python
            from dorsal.file import MetadataReader

            reader = MetadataReader()

            try:
                result = reader.index_file("path/to/important_document.docx")
                if result.success:
                    print("Document indexed successfully!")
            except Exception as e:
                print(f"Failed to index document: {e}")
            ```

        Args:
            file_path (str): The path to the local file to process and index.
            public (bool, optional): If True, the file record will be created as public. Defaults to False.
            skip_cache (bool, optional): If True, forces reprocessing even if cache exists.
            overwrite_cache (bool, optional): If True, updates the cache with new results.

        Returns:
            FileIndexResponse: A response object from the API detailing the
                result of the indexing operation.

        Raises:
            DorsalClientError: If there's an error processing the file locally or an API error
                               during indexing.
            FileNotFoundError: If the file_path does not exist.
            IOError: If the file cannot be read by ModelRunner.
        """
        if self.offline:
            raise DorsalError("Cannot index file: MetadataReader is in OFFLINE mode.")
        logger.debug("Starting file indexing for: %s (Public: %s)", file_path, public)
        file_record = self._get_or_create_record(
            file_path=file_path, skip_cache=skip_cache, overwrite_cache=overwrite_cache
        )

        logger.debug("Indexing file '%s' (hash: %s) to DorsalHub...", file_path, file_record.hash)

        api_response: FileIndexResponse
        if public:
            api_response = self._client.index_public_file_records(file_records=[file_record])
        else:
            api_response = self._client.index_private_file_records(file_records=[file_record])

        log_status_msg = "processed by API"
        if api_response.results:
            result_item = api_response.results[0]
            result_item.file_path = file_path

            was_created = False
            if api_response.response and api_response.response.status_code == 201:
                was_created = True
            if was_created:
                log_status_msg = "newly indexed"
            elif api_response.success > 0:
                log_status_msg = "updated/existing"
            elif api_response.unauthorized > 0:
                log_status_msg = "unauthorized"
            elif api_response.error > 0:
                log_status_msg = "error during indexing attempt"

            logger.debug(
                "File '%s' %s on DorsalHub. URL: %s",
                file_path,
                log_status_msg,
                getattr(result_item, "url", "N/A"),
            )
        elif api_response.error > 0:
            logger.error("Failed to index file %s.", file_path)
        else:
            raw_resp_text = api_response.response.text[:200] if api_response.response else "N/A"
            logger.warning(
                "Unexpected response from server for file %s (no results or errors in structured response, "
                "but success/error counts are Total:%d, Success:%d, Error:%d, Unauthorized:%d). "
                "Raw response snippet: %s",
                file_path,
                api_response.total,
                api_response.success,
                api_response.error,
                api_response.unauthorized,
                raw_resp_text,
            )
        return api_response

    @overload
    def scan_directory(
        self,
        dir_path: str,
        *,
        recursive: bool = False,
        return_errors: Literal[False] = False,
        show_progress: bool | None = None,
        console: Console | None = None,
        palette: dict | None = None,
        skip_cache: bool = False,
        overwrite_cache: bool = False,
        follow_symlinks: bool = True,
        lazy: bool = False,
    ) -> list[LocalFile]: ...

    @overload
    def scan_directory(
        self,
        dir_path: str,
        *,
        recursive: bool = False,
        return_errors: Literal[True],
        show_progress: bool | None = None,
        console: Console | None = None,
        palette: dict | None = None,
        skip_cache: bool = False,
        overwrite_cache: bool = False,
        follow_symlinks: bool = True,
        lazy: bool = False,
    ) -> tuple[list[LocalFile], list[str]]: ...

    def scan_directory(
        self,
        dir_path: str,
        *,
        recursive: bool = False,
        return_errors: bool = False,
        show_progress: bool | None = None,
        console: Console | None = None,
        palette: dict | None = None,
        skip_cache: bool = False,
        overwrite_cache: bool = False,
        follow_symlinks: bool = True,
        lazy: bool = False,
    ) -> list[LocalFile] | tuple[list[LocalFile], list[str]]:
        """Scans a directory and runs the pipeline on all found files.

        This method discovers all files within a given directory (and optionally
        its subdirectories), runs the metadata extraction pipeline on each one,
        and returns a list of the resulting `LocalFile` objects. This is an
        offline operation.

        This method utilizes the internal `ModelRunner` instance of this
        `MetadataReader` (via `_get_or_create_record`), ensuring that any custom
        pipeline configuration provided during initialization is respected and
        that initialization overhead is incurred only once, not per file.

        Example:
            ```python
            from dorsal.file import MetadataReader

            # Initialize with a custom pipeline if needed
            reader = MetadataReader(model_config="my_pipeline.json")

            files_to_process = reader.scan_directory("path/to/images", recursive=True)

            print(f"Found and processed {len(files_to_process)} image files.")
            ```

        Args:
            dir_path (str): The path to the directory to scan.
            recursive (bool, optional): If True, scans all subdirectories
                recursively. Defaults to False.
            return_errors (bool, optional): If True, returns a tuple containing
                the list of successfully processed files and a list of error strings.
                Defaults to False.
            console (Console, optional): A Rich Console instance for progress bars.
            palette (dict, optional): A dictionary defining colors for the progress bar.
            skip_cache (bool, optional): If True, forces reprocessing even if
                valid cache exists. Defaults to False.
            overwrite_cache (bool, optional): If True, updates the cache with
                new results. Defaults to False.
            follow_symlinks (bool, optional): If True, resolves symlinks to their
                targets. Defaults to True.
            lazy (bool, optional): If True, processes files via an iterator to avoid
                scanning the entire tree before processing begins. Defaults to False.

        Returns:
            list[LocalFile]: initialized `LocalFile` instances.
            Or
            tuple:
            - list[LocalFile]: initialized `LocalFile` instances.
            - list[str]: errors (useful for CLI output/debugging).
        """
        logger.debug("Starting directory read for: %s (Recursive: %s, Lazy: %s)", dir_path, recursive, lazy)
        try:
            file_paths = get_file_paths(dir_path=dir_path, recursive=recursive, lazy=lazy)
        except FileNotFoundError:
            logger.error("Directory not found: %s", dir_path)
            raise
        except Exception as e:
            logger.error("Error scanning directory %s: %s", dir_path, e)
            raise DorsalError(
                message=f"An error occurred while scanning directory: {dir_path}.",
                original_exception=e,
            ) from e

        total_count = None
        if isinstance(file_paths, list):
            if not file_paths:
                logger.debug("No files found to read in directory: %s", dir_path)
                return ([], []) if return_errors else []
            total_count = len(file_paths)
            logger.debug("Found %d file(s). Processing with local models...", total_count)
        else:
            logger.debug("Processing files in LAZY mode (unknown total count)...")

        local_files: list[LocalFile] = []
        warnings: list[str] = []

        should_display = should_show_progress(show_progress, console)

        rich_progress = None
        iterator: Iterable[str]

        if should_display and is_jupyter_environment():
            iterator = tqdm(file_paths, desc="Processing Files", total=total_count)
        elif should_display:
            from dorsal.cli.themes.palettes import DEFAULT_PALETTE

            active_palette = palette if palette is not None else DEFAULT_PALETTE
            progress_columns = (
                SpinnerColumn(),
                TextColumn(
                    "{task.description}",
                    style=active_palette.get("progress_description", "default"),
                ),
                BarColumn(complete_style=active_palette.get("progress_bar", "default")),
                TaskProgressColumn(style=active_palette.get("progress_percentage", "default")),
                MofNCompleteColumn(),
                TimeElapsedColumn(),
                TimeRemainingColumn(),
            )
            rich_progress = Progress(
                *progress_columns,
                console=console,
                transient=True,
                redirect_stdout=True,
                redirect_stderr=True,
            )
            task_id = rich_progress.add_task("Initializing...", total=total_count)
            iterator = file_paths
        else:
            iterator = file_paths

        with rich_progress if rich_progress else open(os.devnull, "w"):
            for file_path in iterator:
                if rich_progress:
                    filename = os.path.basename(file_path)
                    rich_progress.update(task_id, description=f"Processing {filename}...")

                try:
                    processing_path = file_path
                    if follow_symlinks:
                        try:
                            path_obj = pathlib.Path(file_path)
                            resolved = path_obj.resolve()
                            if resolved.exists():
                                processing_path = str(resolved)
                        except (OSError, RuntimeError) as err:
                            logger.debug("Failed to resolve symlink during scan: %s - %s", file_path, err)
                            pass

                    record = self._get_or_create_record(
                        file_path=processing_path,
                        skip_cache=skip_cache,
                        overwrite_cache=overwrite_cache,
                        follow_symlinks=follow_symlinks,
                    )

                    local_file_obj = self._file_class(
                        file_path=file_path,
                        client=self._client_instance,
                        offline=self.offline,
                        _file_record=record,
                    )

                    local_files.append(local_file_obj)
                except Exception as e:
                    msg = f"Skipping '{os.path.basename(file_path)}' due to processing error: {e}"
                    logger.error(msg)
                    warnings.append(msg)

                if rich_progress:
                    rich_progress.advance(task_id)

        logger.debug(
            "Successfully processed %d files into %s objects from directory %s.",
            len(local_files),
            self._file_class.__name__,
            dir_path,
        )

        if return_errors:
            return local_files, warnings
        return local_files

    def scan_file(
        self, file_path: str, *, skip_cache: bool, overwrite_cache: bool = False, follow_symlinks: bool = True
    ) -> LocalFile:
        """Runs the metadata extraction pipeline on a single file.

        This method processes a single local file through the configured
        `ModelRunner` pipeline, generating all its metadata. This is an
        offline operation that does not contact the DorsalHub API.

        Example:
            ```python
            from dorsal.file import MetadataReader

            reader = MetadataReader()
            try:
                local_file = reader.scan_file("path/to/document.pdf")
                print(f"Successfully processed: {local_file.name}")
                if local_file.pdf:
                    print(f"Page Count: {local_file.pdf.page_count}")
            except FileNotFoundError:
                print("The specified file does not exist.")
            ```

        Args:
            file_path (str): The path to the local file to process.

        Returns:
            LocalFile: An initialized `LocalFile` instance containing the rich
                metadata extracted by the pipeline.

        Raises:
            FileNotFoundError: If the file_path does not exist or is not a file.
            IOError: If there are issues reading the file.
            DorsalError: For other errors during model execution

        """
        logger.debug("Starting file read for: %s", file_path)
        try:
            local_file_obj = self._file_class(
                file_path=file_path,
                client=self._client_instance,
                use_cache=not skip_cache,
                overwrite_cache=overwrite_cache,
                offline=self.offline,
                follow_symlinks=follow_symlinks,
            )
        except FileNotFoundError:
            logger.error("File not found for reading: %s", file_path)
            raise
        except IOError as e:
            logger.error("IOError reading file %s: %s", file_path, e)
            raise
        except Exception as e:
            logger.error("Failed to read/process file %s: %s", file_path, e)
            raise DorsalError(
                message=f"Failed to process local file: {file_path}.",
                original_exception=e,
            ) from e

        logger.debug(
            "Successfully processed file '%s' into a %s object (hash: %s).",
            file_path,
            self._file_class.__name__,
            local_file_obj.hash,
        )
        return local_file_obj



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib.metadata
import importlib.resources
import tomllib
from typing import Any, Optional
from unittest.mock import MagicMock, patch

import pytest
from pydantic import ValidationError

from dorsal.api.model import run_or_install_model
from dorsal.common.exceptions import DorsalError, DorsalConfigError
from dorsal.common.validators import CallableImportPath
from dorsal.file.configs.model_runner import ModelRunnerPipelineStep


@pytest.fixture
def mock_resolve_target():
    with patch("dorsal.api.model.resolve_target") as mock:
        yield mock


@pytest.fixture
def mock_is_package_installed():
    with patch("dorsal.api.model.is_package_installed") as mock:
        yield mock


@pytest.fixture
def mock_install_model_target():
    with patch("dorsal.api.model.install_model_target") as mock:
        yield mock


@pytest.fixture
def mock_get_model_pipeline():
    with patch("dorsal.api.model.get_model_pipeline") as mock:
        yield mock


@pytest.fixture
def mock_resolve_pipeline_step_models():
    with patch("dorsal.api.model.resolve_pipeline_step_models") as mock:
        mock.return_value = (MagicMock(__name__="MockAnnotator"), MagicMock(__name__="MockValidator"))
        yield mock


@pytest.fixture
def mock_run_model():
    with patch("dorsal.api.model.run_model") as mock:
        yield mock


@pytest.fixture
def mock_entry_points():
    with patch("importlib.metadata.entry_points") as mock:
        yield mock


@pytest.fixture
def mock_resources_files():
    with patch("importlib.resources.files") as mock:
        yield mock


def test_run_or_install_model_happy_path_existing_pipeline(
    mock_resolve_target,
    mock_is_package_installed,
    mock_get_model_pipeline,
    mock_resolve_pipeline_step_models,
    mock_run_model,
    mock_install_model_target,
):
    """Test running a model that is installed and exists in the active pipeline."""
    target = "dorsal-whisper"
    file_path = "/tmp/audio.mp3"
    package_name = "dorsal-whisper"

    mock_resolve_target.return_value = ("package_name", package_name)
    mock_is_package_installed.return_value = True

    existing_step = ModelRunnerPipelineStep(
        annotation_model=CallableImportPath(module="dorsal_whisper", name="WhisperModel"),
        schema_id="audio/transcription",
        package_name=package_name,
    )
    mock_get_model_pipeline.return_value = [existing_step]

    mock_annotator = MagicMock()
    mock_validator = MagicMock()
    mock_resolve_pipeline_step_models.return_value = (mock_annotator, mock_validator)

    run_or_install_model(target, file_path)

    mock_install_model_target.assert_not_called()

    mock_run_model.assert_called_once_with(
        annotation_model=mock_annotator,
        file_path=file_path,
        schema_id=existing_step.schema_id,
        schema_version=existing_step.schema_version,
        validation_model=mock_validator,
        dependencies=existing_step.dependencies,
        options=existing_step.options,
        ignore_linter_errors=existing_step.ignore_linter_errors,
        progress_callback=None,
    )


def test_run_or_install_model_auto_install_success(
    mock_resolve_target,
    mock_is_package_installed,
    mock_install_model_target,
    mock_get_model_pipeline,
    mock_entry_points,
    mock_resources_files,
    mock_resolve_pipeline_step_models,
    mock_run_model,
):
    """Test auto-installing a missing model, then constructing its step from metadata."""
    target = "dorsalhub/whisper"
    file_path = "/tmp/audio.wav"
    package_name = "dorsal-whisper"
    module_name = "dorsal_whisper"

    mock_resolve_target.return_value = ("registry_id", package_name)
    mock_is_package_installed.side_effect = [False, True]

    mock_get_model_pipeline.return_value = []

    mock_ep = MagicMock()
    mock_ep.dist.name = package_name
    mock_ep.module = module_name
    mock_entry_points.return_value = [mock_ep]

    mock_resource_path = MagicMock()
    mock_resource_path.is_file.return_value = True

    mock_resource_path.read_text.return_value = """
        model_class = "WhisperTranscriber"
        schema_id = "audio/transcription"
        dependencies = []
    """
    mock_resources_files.return_value.__truediv__.return_value = mock_resource_path

    run_or_install_model(target, file_path)

    mock_install_model_target.assert_called_once_with(target)

    call_args = mock_resolve_pipeline_step_models.call_args
    assert call_args is not None
    step_arg = call_args.args[0]

    assert isinstance(step_arg, ModelRunnerPipelineStep)
    assert step_arg.annotation_model.name == "WhisperTranscriber"
    assert step_arg.schema_id == "audio/transcription"
    assert step_arg.package_name == package_name

    mock_run_model.assert_called_once()


def test_run_or_install_model_runtime_options(
    mock_resolve_target,
    mock_is_package_installed,
    mock_get_model_pipeline,
    mock_resolve_pipeline_step_models,
    mock_run_model,
):
    """Test applying runtime options and linter flags."""
    target = "dorsal-ocr"
    mock_resolve_target.return_value = ("package_name", "dorsal-ocr")
    mock_is_package_installed.return_value = True

    base_step = ModelRunnerPipelineStep(
        annotation_model=CallableImportPath(module="m", name="c"),
        schema_id="valid/schema-id",
        options={"lang": "en"},
        package_name="dorsal-ocr",
    )
    mock_get_model_pipeline.return_value = [base_step]

    run_or_install_model(target, "doc.pdf", options={"lang": "fr", "fast": True}, ignore_linter_errors=True)

    call_args = mock_run_model.call_args

    assert call_args.kwargs["options"] == {"lang": "fr", "fast": True}
    assert call_args.kwargs["ignore_linter_errors"] is True


def test_annotator_execution_failure_propagates(
    mock_resolve_target,
    mock_is_package_installed,
    mock_get_model_pipeline,
    mock_resolve_pipeline_step_models,
    mock_run_model,
):
    """Test that exceptions raised during annotation are propagated."""
    mock_resolve_target.return_value = ("package_name", "dorsal-fail")
    mock_is_package_installed.return_value = True

    step = ModelRunnerPipelineStep(
        annotation_model=CallableImportPath(module="m", name="c"),
        schema_id="valid/schema-id",
        package_name="dorsal-fail",
    )
    mock_get_model_pipeline.return_value = [step]

    mock_run_model.side_effect = ValueError("Processing failed")

    with pytest.raises(ValueError, match="Processing failed"):
        run_or_install_model("dorsal-fail", "f.txt")


def test_run_or_install_model_auto_install_failure(
    mock_resolve_target, mock_is_package_installed, mock_install_model_target
):
    target = "owner/broken-repo"
    mock_resolve_target.return_value = ("registry_id", "dorsal-broken")
    mock_is_package_installed.return_value = False
    mock_install_model_target.side_effect = Exception("Network error")

    with pytest.raises(DorsalError, match="Failed to auto-install model"):
        run_or_install_model(target, "file.txt")


def test_construct_step_missing_entry_point(
    mock_resolve_target, mock_is_package_installed, mock_get_model_pipeline, mock_entry_points
):
    target = "dorsal-ghost"
    mock_resolve_target.return_value = ("package_name", "dorsal-ghost")
    mock_is_package_installed.return_value = True
    mock_get_model_pipeline.return_value = []

    mock_ep = MagicMock()
    mock_ep.dist.name = "other-package"
    mock_entry_points.return_value = [mock_ep]

    with pytest.raises(DorsalError, match="does not declare a 'dorsal.models' entry point"):
        run_or_install_model(target, "f.txt")


def test_load_package_config_missing_file(
    mock_resolve_target, mock_is_package_installed, mock_get_model_pipeline, mock_entry_points, mock_resources_files
):
    mock_resolve_target.return_value = ("package_name", "dorsal-pkg")
    mock_is_package_installed.return_value = True
    mock_get_model_pipeline.return_value = []

    mock_ep = MagicMock()
    mock_ep.dist.name = "dorsal-pkg"
    mock_ep.module = "dorsal_pkg"
    mock_entry_points.return_value = [mock_ep]

    mock_path = MagicMock()
    mock_path.is_file.return_value = False
    mock_resources_files.return_value.__truediv__.return_value = mock_path

    with pytest.raises(DorsalConfigError, match="missing 'model_config.toml'"):
        run_or_install_model("dorsal-pkg", "f.txt")



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
from dorsal.session import (
    get_shared_dorsal_client,
    set_shared_dorsal_client,
    clear_shared_dorsal_client,
    get_shared_cache,
    set_shared_cache,
    clear_shared_cache,
    get_metadata_reader,
)
from dorsal.client.dorsal_client import DorsalClient
from dorsal.file.cache.dorsal_cache import DorsalCache


@pytest.fixture(autouse=True)
def clean_globals():
    """Reset global state before and after each test."""
    clear_shared_dorsal_client()
    clear_shared_cache()
    # _METADATA_READER doesn't have a public clear method in the snippet,
    # but it's less stateful.
    yield
    clear_shared_dorsal_client()
    clear_shared_cache()


def test_get_shared_client_init():
    client = get_shared_dorsal_client(api_key="test_key")
    assert isinstance(client, DorsalClient)
    assert client.api_key == "test_key"

    # Second call returns same instance
    client2 = get_shared_dorsal_client()
    assert client2 is client


def test_set_shared_client(mocker):
    mock_client = mocker.MagicMock(spec=DorsalClient)
    set_shared_dorsal_client(mock_client)

    assert get_shared_dorsal_client() is mock_client


def test_get_shared_cache_init():
    cache = get_shared_cache()
    assert isinstance(cache, DorsalCache)

    cache2 = get_shared_cache()
    assert cache2 is cache


def test_set_shared_cache(mocker):
    mock_cache = mocker.MagicMock(spec=DorsalCache)
    set_shared_cache(mock_cache)
    assert get_shared_cache() is mock_cache


def test_clear_shared_cache(mocker):
    mock_cache = mocker.MagicMock(spec=DorsalCache)
    set_shared_cache(mock_cache)

    clear_shared_cache()
    mock_cache.close.assert_called_once()

    # Next get should create new one
    assert get_shared_cache() is not mock_cache


def test_get_metadata_reader():
    # Just verify it returns the object and singleton behavior
    reader = get_metadata_reader()
    assert reader is not None
    assert get_metadata_reader() is reader



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from typer.testing import CliRunner
from unittest.mock import MagicMock
from rich.panel import Panel


from dorsal.cli import app
from dorsal.common.exceptions import DorsalClientError

runner = CliRunner()
COLLECTION_ID = "col_abc123"


@pytest.fixture
def mock_update_collection_cmd(mocker):
    """Mocks dependencies for the `collection update` command."""
    # Mock the Pydantic model that is returned on a successful update
    mock_updated_collection = MagicMock()
    mock_updated_collection.name = "My Updated Collection Name"

    # Patch the backend API function at its source due to lazy loading
    mock_update = mocker.patch("dorsal.api.collection.update_collection", return_value=mock_updated_collection)
    return mock_update


def test_update_collection_name_and_desc(mock_rich_console, mock_update_collection_cmd):
    """Tests updating both name and description successfully."""
    result = runner.invoke(
        app,
        [
            "collection",
            "update",
            COLLECTION_ID,
            "--name",
            "New Name",
            "--description",
            "New Desc",
        ],
    )

    assert result.exit_code == 0
    mock_update_collection_cmd.assert_called_once_with(
        collection_id=COLLECTION_ID,
        name="New Name",
        description="New Desc",
        mode="pydantic",
    )

    # Check that a success panel was printed
    printed_object = mock_rich_console.print.call_args.args[0]
    assert isinstance(printed_object, Panel)
    assert "Update Complete" in str(printed_object.title)
    assert "updated successfully" in str(printed_object.renderable)


def test_update_collection_only_name(mock_rich_console, mock_update_collection_cmd):
    """Tests updating only the collection name."""
    runner.invoke(app, ["collection", "update", COLLECTION_ID, "--name", "Only The Name"])

    mock_update_collection_cmd.assert_called_once_with(
        collection_id=COLLECTION_ID,
        name="Only The Name",
        description=None,
        mode="pydantic",
    )


def test_update_collection_only_desc(mock_rich_console, mock_update_collection_cmd):
    """Tests updating only the collection description."""
    runner.invoke(app, ["collection", "update", COLLECTION_ID, "--description", "Only The Desc"])

    mock_update_collection_cmd.assert_called_once_with(
        collection_id=COLLECTION_ID,
        name=None,
        description="Only The Desc",
        mode="pydantic",
    )


def test_update_collection_no_options_error(mock_update_collection_cmd):
    """Tests that providing no update options results in a ValueError from the API."""
    error_msg = "At least one field (name or description) must be provided to update."
    mock_update_collection_cmd.side_effect = ValueError(error_msg)

    result = runner.invoke(app, ["collection", "update", COLLECTION_ID])

    assert result.exit_code != 0
    assert error_msg in result.output


def test_update_collection_api_error(mock_update_collection_cmd):
    """Tests graceful failure on a DorsalClientError."""
    mock_update_collection_cmd.side_effect = DorsalClientError("Collection not found")

    result = runner.invoke(app, ["collection", "update", COLLECTION_ID, "--name", "New Name"])

    assert result.exit_code != 0
    assert "API Error: Collection not found" in result.output



# Copyright 2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import sys
import shutil
from typing import Any, Dict

from rich.panel import Panel
from rich.prompt import Confirm
from rich.text import Text
import typer


logger = logging.getLogger(__name__)


def check_and_confirm_model_install(
    target: str, palette: Dict[str, str], force: bool = False, yes: bool = False
) -> None:
    """
    Checks if a model target is safe to install and prompts the user for confirmation.

    Args:
        target: The model identifier (registry ID, git URL, etc).
        palette: The CLI color palette.
        force: If True, skips checks (assumes user knows what they are doing).
        yes: If True, skips interactive confirmation.
    """
    from dorsal.common.constants import WEB_URL
    from dorsal.common.exceptions import AuthError
    from dorsal.common.cli import exit_cli, get_error_console
    from dorsal.session import get_shared_dorsal_client
    from dorsal.registry.validators import is_registry_id

    if force or yes:
        return

    error_console = get_error_console()

    if "pipx" in sys.prefix:
        error_console.print(
            f"[{palette.get('info', 'dim')}]Note: You are running inside a pipx environment.\n"
            "The model will be available to the CLI, but NOT to external Python scripts.[/]"
        )

    display_meta = {"Model": target, "Source": "Local Path"}
    status_badge = f"[{palette.get('warning', 'bold yellow')}]Unverified[/]"
    border_style = palette.get("panel_border_warning", "yellow")
    link_style = palette.get("link", "blue underline")

    if is_registry_id(target):
        try:
            with error_console.status(f"[{palette.get('info', 'dim')}]Fetching model details...[/]"):
                client = get_shared_dorsal_client()
                reg_data = client.get_registry_model(target)

                if reg_data.is_official or reg_data.is_verified:
                    status_badge = f"[{palette.get('panel_border', 'bold blue')}]Verified[/]"
                    border_style = palette.get("panel_border", "blue")

                dorsalhub_url = f"{WEB_URL}/models/{reg_data.namespace}/{reg_data.name}"
                display_meta = {
                    "Model": f"{reg_data.namespace}/{reg_data.name}",
                    "Status": status_badge,
                    "URL": f"[{link_style} link={dorsalhub_url}]{dorsalhub_url}[/]",
                    "Description": reg_data.description or "No description provided.",
                }

                if reg_data.install_url:
                    if reg_data.install_url.startswith("git+") and not shutil.which("git"):
                        _handle_missing_git(palette)

                    raw_url = reg_data.install_url.replace("git+", "").split("@")[0]
                    display_meta["Source Code"] = f"[{link_style} link={raw_url}]{raw_url}[/]"

                if reg_data.created_at:
                    display_meta["Published"] = reg_data.created_at.date().isoformat()

        except AuthError:
            raise
        except typer.Exit:
            raise
        except Exception as e:
            logger.debug(f"Metadata fetch failed: {e}")
            if "/" in target:
                _handle_registry_error(target, e, palette)
            display_meta["Warning"] = "Could not fetch remote metadata."

    msg_lines = []
    for k, v in display_meta.items():
        msg_lines.append(f"[{palette.get('key', 'dim')}]{k}:[/] {v}")

    if "Unverified" in status_badge:
        msg_lines.append("\nYou are about to install executable code from an unverified source.")
        msg_lines.append("Please review the source above before proceeding.")

    error_console.print(
        Panel(
            "\n".join(msg_lines),
            title=f"[{palette.get('panel_title_warning', 'bold yellow')}]Model Info[/]",
            border_style=border_style,
            expand=False,
        )
    )

    if not Confirm.ask("Do you trust this source and want to proceed?", console=error_console):
        error_console.print(f"[{palette.get('error', 'bold red')}]Cancelled.[/]")
        exit_cli(code=0)


def _handle_missing_git(palette: Dict[str, str]):
    from dorsal.common.cli import exit_cli, EXIT_CODE_ERROR, get_error_console

    error_console = get_error_console()
    message = Text.assemble(
        ("This model requires Git to install.\n\n", "default"),
        ("Dorsal could not find the ", "default"),
        ("git", f"bold {palette.get('primary_value', 'cyan')}"),
        (" command in your system PATH.\n\n", "default"),
        ("To proceed, please install Git from:\n", "default"),
        ("https://git-scm.com/downloads", palette.get("link", "blue underline")),
    )
    error_console.print(
        Panel(
            message,
            expand=False,
            title=f"[{palette.get('panel_title_error', 'bold red')}]Missing System Dependency[/]",
            border_style=palette.get("panel_border_error", "red"),
        )
    )
    exit_cli(code=EXIT_CODE_ERROR)


def _handle_registry_error(target: str, e: Exception, palette: Dict[str, str]):
    from dorsal.common.exceptions import NotFoundError
    from dorsal.common.cli import exit_cli, EXIT_CODE_ERROR, get_error_console

    error_console = get_error_console()
    if isinstance(e, NotFoundError) or "404" in str(e):
        error_console.print(f"[{palette.get('error', 'bold red')}]Error: Model '{target}' not found in registry.[/]")
    else:
        error_console.print(
            f"[{palette.get('error', 'bold red')}]Error: Failed to connect to registry for '{target}'.[/]\n{e}"
        )
    exit_cli(code=EXIT_CODE_ERROR)



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import logging
import subprocess
from pathlib import Path
from unittest.mock import MagicMock, mock_open
from typer.testing import CliRunner
from rich.console import Console
from rich.text import Text
from rich.panel import Panel


from dorsal.cli import app
from dorsal.common.exceptions import AuthError, NetworkError


import typer
from rich.logging import RichHandler

app_repro = typer.Typer()


@app_repro.callback()
def main_callback():
    console = Console(stderr=True)
    logging.basicConfig(
        level=logging.INFO,
        format="%(message)s",
        handlers=[RichHandler(console=console, rich_tracebacks=True)],
    )


@app_repro.command()
def hello():
    logging.info("Hello from command")
    raise typer.Exit()


runner = CliRunner()

MOCK_USER_INFO = {
    "user_id": "usr_12345",
    "name": "Test User",
    "email": "test@example.com",
    "account_status": "active",
}


def test_login_success(mocker, mock_rich_console, mock_auth_app):
    """Tests a successful 'auth login' (Global scope)."""
    mocker.patch("dorsal.client.DorsalClient").return_value.verify_credentials.return_value = MOCK_USER_INFO
    mocker.patch("dorsal.common.config.get_global_config_path", return_value=Path("/fake/path/dorsal.toml"))

    result = runner.invoke(app, ["auth", "login", "--apikey", "test-key-123"])

    assert result.exit_code == 0
    final_message = str(mock_rich_console.print.call_args_list[3].args[0])
    assert "Your API key has been saved to" in final_message
    assert str(Path("/fake/path/dorsal.toml")) in final_message


def test_login_project_scope(mocker, mock_rich_console, mock_auth_app):
    """
    Tests 'auth login --project'.
    Covers the logic that warns the user about committing the config file.
    """
    mocker.patch("dorsal.client.DorsalClient").return_value.verify_credentials.return_value = MOCK_USER_INFO

    # Mock finding a local dorsal.toml
    project_path = Path("/my/project/dorsal.toml")
    mocker.patch("dorsal.common.config.find_project_config_path", return_value=project_path)
    mock_write = mocker.patch("dorsal.common.auth.write_auth_config")

    result = runner.invoke(app, ["auth", "login", "--apikey", "pk_123", "--project"])

    assert result.exit_code == 0

    # Verify we wrote to project scope
    mock_write.assert_called_with(api_key="pk_123", email="test@example.com", scope="project")

    # Verify the security warning panel was printed
    # It's usually the last thing printed
    last_print_arg = mock_rich_console.print.call_args[0][0]
    assert isinstance(last_print_arg, Panel)
    assert "Action Required" in str(last_print_arg.title)
    assert "must not" in str(last_print_arg.renderable)
    assert "dorsal auth gitignore" in str(last_print_arg.renderable)


# --- Logout Tests ---


def test_logout_success(mocker, mock_rich_console, mock_auth_app):
    """Tests a successful 'auth logout'."""
    from dorsal.common.auth import APIKeySource

    mocker.patch(
        "dorsal.common.auth.get_api_key_details", return_value={"source": APIKeySource.PROJECT, "value": "a-real-key"}
    )
    mocker.patch("dorsal.common.auth.remove_api_key", return_value=True)

    result = runner.invoke(app, ["auth", "logout"])

    assert result.exit_code == 0
    printed_text = mock_rich_console.print.call_args.args[0]
    assert "✅ You have been successfully logged out." in str(printed_text)


def test_logout_not_logged_in(mocker, mock_rich_console, mock_auth_app):
    """Tests 'auth logout' when no key exists."""
    from dorsal.common.auth import APIKeySource

    mocker.patch("dorsal.common.auth.get_api_key_details", return_value={"source": APIKeySource.NONE, "value": None})

    result = runner.invoke(app, ["auth", "logout"])
    assert result.exit_code == 0
    assert "not currently logged in" in str(mock_rich_console.print.call_args.args[0])


def test_logout_with_env_var_warning(mocker, mock_rich_console, mock_auth_app):
    """Tests 'auth logout' with an active ENV VAR."""
    from dorsal.common.auth import APIKeySource

    mocker.patch(
        "dorsal.common.auth.get_api_key_details", return_value={"source": APIKeySource.ENV, "value": "env-key"}
    )

    result = runner.invoke(app, ["auth", "logout"])

    assert result.exit_code == 0
    panel_output = mock_rich_console.print.call_args_list[0].args[0]
    assert isinstance(panel_output, Panel)
    assert "Environment Variable Active" in str(panel_output.title)


# --- Whoami Tests ---


def test_whoami_success(mocker, mock_rich_console, mock_auth_app):
    """Tests 'auth whoami' success path."""
    mocker.patch(
        "dorsal.session.get_shared_dorsal_client"
    ).return_value.verify_credentials.return_value = MOCK_USER_INFO

    result = runner.invoke(app, ["auth", "whoami"])

    assert result.exit_code == 0
    panel_output = mock_rich_console.print.call_args_list[2].args[0]
    assert isinstance(panel_output, Panel)
    assert "👤 Authenticated User" in str(panel_output.title)
    assert "Test User" in str(panel_output.renderable)


def test_whoami_json(mocker, mock_rich_console, mock_auth_app):
    """Tests 'auth whoami --json'."""
    mocker.patch(
        "dorsal.session.get_shared_dorsal_client"
    ).return_value.verify_credentials.return_value = MOCK_USER_INFO

    result = runner.invoke(app, ["auth", "whoami", "--json"])

    assert result.exit_code == 0
    # Verify the output was raw JSON (not a panel)
    output_str = mock_rich_console.print.call_args.args[0]
    assert '"user_id": "usr_12345"' in output_str
    assert '"email": "test@example.com"' in output_str


def test_whoami_auth_error(mocker, mock_auth_app):
    """
    Tests that the command bubbles up AuthError (does not swallow it).
    """
    # 1. Setup failure
    mocker.patch("dorsal.session.get_shared_dorsal_client").side_effect = AuthError("Auth failed.")

    # 2. Run command
    result = runner.invoke(app, ["auth", "whoami"])

    # 3. Assert CRASH (Not 0) and correct Exception type
    assert result.exit_code != 0
    assert isinstance(result.exception, AuthError)


# --- Gitignore Tests (High Coverage Gain) ---


def test_gitignore_no_git(mocker, mock_rich_console):
    """Tests behavior when 'git' is not installed."""
    mocker.patch("shutil.which", return_value=None)

    result = runner.invoke(app, ["auth", "gitignore"])

    assert result.exit_code != 0
    assert "'git' command not found" in str(mock_rich_console.print.call_args.args[0])


def test_gitignore_not_repo(mocker, mock_rich_console):
    """Tests behavior when not in a git repository."""
    mocker.patch("shutil.which", return_value="/usr/bin/git")
    mocker.patch("subprocess.run", side_effect=subprocess.CalledProcessError(128, "git"))

    result = runner.invoke(app, ["auth", "gitignore"])

    assert result.exit_code == 0
    assert "does not appear to be a Git repository" in str(mock_rich_console.print.call_args.args[0])


def test_gitignore_no_config(mocker, mock_rich_console):
    """Tests behavior when no dorsal.toml is found."""
    mocker.patch("shutil.which", return_value=True)
    mocker.patch("subprocess.run").return_value.stdout = "/repo/root"
    mocker.patch("dorsal.common.config.get_project_level_config", return_value=(None, None))

    result = runner.invoke(app, ["auth", "gitignore"])

    assert result.exit_code == 0
    assert "No project-level config file found" in str(mock_rich_console.print.call_args.args[0])


def test_gitignore_config_safe(mocker, mock_rich_console):
    """Tests behavior when config exists but has no auth section."""
    mocker.patch("shutil.which", return_value=True)
    mocker.patch("subprocess.run").return_value.stdout = "/repo/root"

    # Config exists, is NOT empty, but has NO auth section
    config_path = Path("/repo/root/dorsal.toml")
    safe_config_data = {"general": {"theme": "dark"}}
    mocker.patch("dorsal.common.config.get_project_level_config", return_value=(safe_config_data, config_path))

    result = runner.invoke(app, ["auth", "gitignore"])

    assert result.exit_code == 0

    all_output = "".join(str(call.args[0]) for call in mock_rich_console.print.call_args_list)
    assert "contains no 'auth' section" in all_output


def test_gitignore_already_exists(mocker, mock_rich_console):
    """Tests behavior when config is already in .gitignore."""
    mocker.patch("shutil.which", return_value=True)
    mocker.patch("subprocess.run").return_value.stdout = "/repo/root"

    config_path = Path("/repo/root/dorsal.toml")
    mocker.patch("dorsal.common.config.get_project_level_config", return_value=({"auth": "secrets"}, config_path))

    # Mock reading .gitignore
    mock_file = mock_open(read_data="random_file.txt\n/dorsal.toml\n")
    mocker.patch("builtins.open", mock_file)
    mocker.patch("pathlib.Path.exists", return_value=True)

    result = runner.invoke(app, ["auth", "gitignore"])

    assert result.exit_code == 0
    assert "already exists in .gitignore" in str(mock_rich_console.print.call_args_list[-2].args[0])
    # Should print the "File Still Tracked?" panel
    assert "File Still Tracked?" in str(mock_rich_console.print.call_args_list[-1].args[0].title)


def test_gitignore_add_success(mocker, mock_rich_console):
    """Tests successfully adding the config to .gitignore."""
    mocker.patch("shutil.which", return_value=True)
    mocker.patch("subprocess.run").return_value.stdout = "/repo/root"

    config_path = Path("/repo/root/dorsal.toml")
    mocker.patch("dorsal.common.config.get_project_level_config", return_value=({"auth": "secrets"}, config_path))

    # Mock reading .gitignore (empty) and checking existence
    mock_file = mock_open(read_data="")
    mocker.patch("builtins.open", mock_file)
    mocker.patch("pathlib.Path.exists", return_value=True)

    # Force user confirmation "y"
    mocker.patch("typer.confirm", return_value=True)

    result = runner.invoke(app, ["auth", "gitignore"])

    assert result.exit_code == 0

    # Check it wrote to the file
    mock_file().write.assert_called()
    written_data = mock_file().write.call_args[0][0]
    assert "/dorsal.toml" in written_data

    # Verify success message
    assert "Successfully added" in str(mock_rich_console.print.call_args.args[0])


def test_final_diagnostic_with_caplog(caplog):
    """Runs diagnostic app and captures log."""
    with caplog.at_level(logging.INFO):
        result = runner.invoke(app_repro, ["hello"])
    assert result.exit_code == 0
    assert "Hello from command" in caplog.text



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import pathlib
from typer.testing import CliRunner
from unittest.mock import MagicMock


from dorsal.cli import app
from dorsal.cli.cache_app import export_cmd

runner = CliRunner()


@pytest.fixture
def mock_export_cache_cmd(mocker):
    """Mocks dependencies for the `cache export` command."""
    # Patch the backend function at its original source location
    mock_export = mocker.patch("dorsal.file.utils.cache.export_cache", return_value=1234)

    # Mock the constants directory to avoid filesystem interaction
    mock_exports_dir = MagicMock(spec=pathlib.Path)
    mocker.patch("dorsal.common.constants.CLI_EXPORTS_DIR", mock_exports_dir)

    # Mock datetime to create a predictable timestamped filename
    mock_dt = MagicMock()
    mock_dt.now.return_value.strftime.return_value = "20250101-120000"
    mocker.patch.object(export_cmd, "datetime", mock_dt)

    return {
        "export_cache": mock_export,
        "exports_dir": mock_exports_dir,
    }


def test_export_cache_default_path_and_format(mock_rich_console, mock_export_cache_cmd):
    """Tests exporting with default options, which should create a timestamped filename."""
    result = runner.invoke(app, ["cache", "export"])

    assert result.exit_code == 0
    mock_export_cache_cmd["exports_dir"].mkdir.assert_called_once_with(parents=True, exist_ok=True)

    expected_path = mock_export_cache_cmd["exports_dir"] / "cache-export-20250101-120000.json.gz"

    mock_export_cache_cmd["export_cache"].assert_called_once_with(
        output_path=expected_path, format="json.gz", include_records=True
    )

    all_printed_text = "".join(str(call.args[0]) for call in mock_rich_console.print.call_args_list)
    assert f"Successfully exported 1,234 records to '{expected_path}'" in all_printed_text


def test_export_cache_with_output_path_inferred_format(mock_rich_console, mock_export_cache_cmd, tmp_path):
    """Tests exporting to a specific path and ensuring the format is inferred from the extension."""
    output_file = tmp_path / "my-cache.json"
    result = runner.invoke(app, ["cache", "export", "--output", str(output_file)])

    assert result.exit_code == 0
    mock_export_cache_cmd["export_cache"].assert_called_once_with(
        output_path=output_file.resolve(), format="json", include_records=True
    )

    all_printed_text = "".join(str(call.args[0]) for call in mock_rich_console.print.call_args_list)
    assert f"Successfully exported 1,234 records to '{output_file.resolve()}'" in all_printed_text


def test_export_cache_with_format_override(mock_rich_console, mock_export_cache_cmd, tmp_path):
    """Tests that the --format flag correctly overrides the file extension."""
    output_file = tmp_path / "my-cache.data"
    result = runner.invoke(app, ["cache", "export", "--output", str(output_file), "--format", "json.gz"])

    assert result.exit_code == 0
    mock_export_cache_cmd["export_cache"].assert_called_once_with(
        output_path=output_file.resolve(), format="json.gz", include_records=True
    )


def test_export_cache_no_records(mock_export_cache_cmd):
    """Tests that the --no-records flag is passed correctly to the backend."""
    result = runner.invoke(app, ["cache", "export", "--no-records"])

    assert result.exit_code == 0
    # Check that the final call was made with include_records=False
    assert mock_export_cache_cmd["export_cache"].call_args.kwargs["include_records"] is False


def test_export_cache_invalid_format():
    """Tests that providing an unsupported format causes a graceful failure."""
    result = runner.invoke(app, ["cache", "export", "--format", "xml"])

    assert result.exit_code != 0
    assert "Invalid format 'xml'" in result.output


def test_export_cache_io_error(mock_export_cache_cmd):
    """Tests that an IOError from the backend is handled gracefully."""
    mock_export_cache_cmd["export_cache"].side_effect = IOError("Disk is full")

    result = runner.invoke(app, ["cache", "export"])

    assert result.exit_code != 0
    assert "Export failed: Disk is full" in result.output



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import logging
from dorsal.file.utils import get_quick_hash, multi_hash
from dorsal.common.exceptions import QuickHashFileSizeError, QuickHashFileInstabilityError


# --- Fixtures ---
@pytest.fixture
def mock_file_hasher(mocker):
    return mocker.patch("dorsal.file.utils.FILE_HASHER")


@pytest.fixture
def mock_quick_hasher(mocker):
    return mocker.patch("dorsal.file.utils.QUICK_HASHER")


@pytest.fixture
def mock_filesize(mocker):
    return mocker.patch("dorsal.file.utils.get_filesize", return_value=1000)


@pytest.fixture
def mock_os_path_getsize(mocker):
    return mocker.patch("os.path.getsize", return_value=1000)


# --- get_quick_hash Tests ---


def test_get_quick_hash_success(mock_quick_hasher, mock_filesize):
    mock_quick_hasher.hash.return_value = "qh123"
    assert get_quick_hash("file.txt") == "qh123"
    mock_quick_hasher.hash.assert_called_with(file_path="file.txt", file_size=1000, follow_symlinks=True)


def test_get_quick_hash_fallback(mock_quick_hasher, mock_filesize, mocker):
    """Test fallback to SHA256 when quick hash returns None."""
    mock_quick_hasher.hash.return_value = None
    mocker.patch("dorsal.file.utils.FILE_HASHER.hash_sha256", return_value="sha123")

    assert get_quick_hash("file.txt", fallback_to_sha256=True) == "sha123"


def test_get_quick_hash_os_error(mock_filesize):
    mock_filesize.side_effect = OSError("Disk error")
    with pytest.raises(OSError):
        get_quick_hash("file.txt")


# --- multi_hash Tests ---


def test_multi_hash_success(mock_file_hasher, mock_quick_hasher, mock_os_path_getsize):
    """Test full success path."""
    # Setup returns
    mock_file_hasher.hash.return_value = {"SHA-256": "sha", "BLAKE3": "blake"}
    mock_quick_hasher.hash.return_value = "quick"

    result = multi_hash("file.txt", similarity_hash=True)

    assert result["SHA-256"] == "sha"
    assert result["BLAKE3"] == "blake"
    assert result["QUICK"] == "quick"

    # Verify calculate_tlsh was passed through
    call_kwargs = mock_file_hasher.hash.call_args[1]
    assert call_kwargs["calculate_tlsh"] is True


def test_multi_hash_quick_fail_graceful(mock_file_hasher, mock_quick_hasher, mock_os_path_getsize, caplog):
    """Test that QuickHash failure doesn't crash the whole operation."""
    mock_file_hasher.hash.return_value = {"SHA-256": "sha"}
    mock_quick_hasher.hash.side_effect = QuickHashFileInstabilityError("Changed")

    result = multi_hash("file.txt")

    assert "SHA-256" in result
    assert "QUICK" not in result
    assert "QuickHash generation for 'file.txt' failed" in caplog.text


def test_multi_hash_main_hasher_fail(mock_file_hasher, mock_os_path_getsize):
    """Test that main hasher failure DOES crash the operation."""
    mock_file_hasher.hash.side_effect = OSError("Read fail")

    with pytest.raises(OSError):
        multi_hash("file.txt")


def test_multi_hash_getsize_fail(mock_os_path_getsize):
    mock_os_path_getsize.side_effect = OSError("Stat fail")
    with pytest.raises(OSError):
        multi_hash("file.txt")



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rich.panel import Panel
from rich.table import Table
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from dorsal.file.validators.collection import FileCollection


def collection_metadata(collection: "FileCollection", palette: dict[str, str]) -> Panel:
    """Creates a rich Panel with the collection's metadata."""
    from dorsal.file.utils.size import human_filesize

    if collection.date_modified is not None:
        date_modified = collection.date_modified.strftime("%Y-%m-%d %H:%M:%S")
    else:
        date_modified = "None"

    metadata_table = Table.grid(expand=False, padding=(0, 1))
    metadata_table.add_column(style=palette.get("key"), width=15)
    metadata_table.add_column(style=palette.get("primary_value"))

    metadata_table.add_row("ID:", collection.collection_id)
    metadata_table.add_row("Name:", collection.name)
    if collection.description:
        metadata_table.add_row("Description:", collection.description)
    metadata_table.add_row("File Count:", f"{collection.file_count:,}")
    metadata_table.add_row("Total Size:", human_filesize(collection.total_size))
    metadata_table.add_row("Access:", "Private" if collection.is_private else "Public")
    metadata_table.add_row("Modified:", date_modified)

    return Panel(
        metadata_table,
        title=f"[{palette.get('panel_title', 'default')}]Collection Metadata[/]",
        border_style=palette.get("panel_border", "default"),
        expand=False,
    )
